import json

from micro_framework.dependencies import Dependency, RunnerDependency
from micro_framework.rpc.connectors import RPCConnector
from .formatters import format_rpc_command
from .proxies import RPCProxy


class RPCDependencyMixin:
    """
    Mixin with methods to communicate with RPCManager and exposing the
    available commands.



    """

    def parse_response(self, response):
        """
        Parses the response from the RPC Manager.
        :param str response: RPC Manager Response String
        :return dict|list: Response Data.
        """
        if isinstance(response, str):
            return json.loads(response).get('data', None)
        return response

    def make_call(self, message, wait_response=True):
        """
        Call the RPC Manager using the instantiated client.

        :param dict|str message: Message to be Sent.
        :param bool wait_response: If True, then we wait for a reply.
        :return json|list: Manager Response
        """
        if isinstance(message, dict):
            message = json.dumps(message)
        connection = self.get_connector().get_connection()
        if wait_response:
            return self.parse_response(connection.send_and_receive(message))

        # Otherwise
        connection.send(message)

    def list_targets(self):
        """
        Call __list_targets__ command. As the name suggests, it retrieves all
        registered RPC targets in the connected server.
        :return list: List of Targets
        """
        message = format_rpc_command('__list_targets__')
        response = self.make_call(message)
        return response

    def get_target(self, target):
        """
        Call __get_target__(target) command.

        It returns a single target information.

        :param str target: The target name.
        :return dict: Target information
        """
        message = format_rpc_command('__get_target__', target)
        response = self.make_call(message)
        return response


class RPCDependency(RPCDependencyMixin, RunnerDependency):
    """
    Dependency Provider that injects a RPCProxy into the worker's target.
    """
    proxy_class: RPCProxy = RPCProxy
    connector_class: RPCConnector = None

    def get_connector_kwargs(self):
        """
        Key arguments to be passed to the connector init method
        :return dict:
        """
        return {}

    def get_connector(self) -> RPCConnector:
        return self.connector_class(**self.get_connector_kwargs())

    def get_proxy_class(self):
        return self.proxy_class

    def get_proxy_kws(self):
        return {
            "connector": self.get_connector(),
            "targets": self.list_targets() or [],
            "name": None
        }

    def get_proxy(self):
        return self.get_proxy_class()(**self.get_proxy_kws())

    async def get_dependency(self, worker):
        # Run in executor because the RPCProxy is not an async class and
        # will block the event loop.
        return await self.runner.event_loop.run_in_executor(
            None, self.get_proxy
        )



import logging
import signal


from loky.process_executor import ProcessPoolExecutor, BrokenProcessPool

from micro_framework.spawners.base import Spawner
from micro_framework.exceptions import BrokenSpawner


logger = logging.getLogger(__name__)


def process_initializer():
    signal.signal(signal.SIGINT, signal.SIG_IGN)


class ProcessPoolSpawner(Spawner, ProcessPoolExecutor):
    def __init__(self, config):
        timeout = config.get("PROCESS_WORKER_TIMEOUT")
        super(ProcessPoolSpawner, self).__init__(
            max_workers=config.get("MAX_WORKERS"),
            timeout=timeout,
            initializer=process_initializer
        )

    def submit(self, *args, **kwargs):
        try:
            return super(ProcessPoolSpawner, self).submit(*args, **kwargs)
        except BrokenProcessPool as exc:
            raise BrokenSpawner from exc

    def shutdown(self, **kwargs) -> None:
        logger.info(
            "Process Worker shutdown initiated, wait until all pending tasks "
            "are consumed."
        )
        super(ProcessPoolSpawner, self).shutdown(**kwargs)



from collections import UserDict


DEFAULT_CONFIG = dict(
    MAX_WORKERS=3,
    WORKER_MODE='thread',
    AMPQ_URI='amqp://guest:guest@localhost:5672',
    ENABLE_METRICS=False,  # This is Preferable to avoid any port collision
    METRICS={
        'HOST': '0.0.0.0',
        'PORT': 8000,
    },
    PROCESS_WORKER_TIMEOUT=10,
    # After 10 seconds the worker will be turned off.
    WEBSOCKET_IP='0.0.0.0',
    WEBSOCKET_PORT=8765,
)


class FrameworkConfig(UserDict):
    """
    Exposes the user configurations but also return the System Default
    configurations when a config is set by the user
    """

    def _deep_update(self, config, default):
        merged_dict = default.copy()
        for name, val in config.items():
            if name in default and isinstance(val, dict):
                merged_val = self._deep_update(val, default[name])
                merged_dict[name] = merged_val
            else:
                merged_dict[name] = val
        return merged_dict

    def __init__(self, config, default_config=None):
        if default_config is None:
            default_config = DEFAULT_CONFIG
        final_config = self._deep_update(config, default_config)
        super(FrameworkConfig, self).__init__(final_config)



import asyncio
import logging

from micro_framework.rpc import RPCManagerMixin, format_rpc_response
from micro_framework.websocket.metrics import active_connections
from micro_framework.websocket.server import WebSocketServer

logger = logging.getLogger(__name__)


class WebSocketManager(RPCManagerMixin, WebSocketServer):
    """
    RPCManager implementation using WebSockets.

    It starts a WebSocketServer that listens to connections and handles
    messages using the format expected by RPCManagerMixin.

    """
    def __init__(self):
        super(WebSocketManager, self).__init__()
        self.started = False

    singleton = True
    context_singleton = False

    async def setup(self):
        logger.debug("Setup WebSocket Manager.")

        self.ip = self.runner.config['WEBSOCKET_IP']
        self.port = self.runner.config["WEBSOCKET_PORT"]

    async def start(self):
        if not self.started:
            logger.debug("Starting WebSocket Manager.")
            await self.runner.spawn_extension(
                self, self.serve, self.ip, self.port
            )
            self.started = True

    async def register_client(self, websocket):
        active_connections.inc()
        await super(WebSocketManager, self).register_client(websocket)
    
    async def unregister_client(self, websocket):
        active_connections.dec()
        await super(WebSocketManager, self).unregister_client(websocket)
    
    async def message_received(self, websocket, message):
        response = await self.consume_message(websocket, message)
        if response:
            await self.send(websocket, response)

    async def send_to_client(self, websocket, data, exception=None):
        """
        Send back to the client outside the event-loop.
        This method is intended to be used by the entrypoint when called
        after a worker has finished.

        :param websocket: Client Connection
        :param message: Response
        :return:
        """
        message = format_rpc_response(data, exception)
        await self.send(websocket, message)



import inspect
import logging

logger = logging.getLogger(__name__)


class Extension:
    """
    An element to be bind into the Service Runner.
    It also binds sub-extensions (Extension inside another)
    """
    _runner = None
    _params = None
    context_singleton = False # Unique to the RunnerContext.
    singleton = False # Unique to the whole service.
    parent = None
    _extensions = set()

    # Define this to True so we won't store the runner nor the parent. Also
    # we won't allow any internal extension to be non-picklable when this is
    # True.
    picklable = False

    def __new__(cls, *args, **kwargs):
        # Hack from Nameko's Extension to enable us to instantiate a new
        # Extension on Bind independent of the arguments it has on __init__
        inst = super(Extension, cls).__new__(cls)
        inst._params = (args, kwargs)

        # This could be on init, but since i'm using __new__ already...
        inst._runner = None
        inst.parent = None
        inst._extensions = set()

        return inst

    def bind_new_extension(self, attr_name, extension, runner, parent):
        binded_ext = extension.bind(runner, parent=parent)
        parent._extensions.add(binded_ext)
        try:
            setattr(parent, attr_name, binded_ext)
        except AttributeError:
            pass
        return binded_ext

    def bind_inner_extensions(self):
        """
        Binds to the extensions declared inside this class.

        # Note for Future me: This is already called in the binded extension.
        No need to use a ext variable here.
        """
        # Retrieve the extensions declared inside this extension
        inner_extensions = inspect.getmembers(
            self, lambda x: isinstance(x, Extension) and x._runner is None
        )
        for attr_name, extension in inner_extensions:
            extension = getattr(self, attr_name)
            if attr_name == "parent" or attr_name == "runner":
                continue
            if self.picklable and not extension.picklable:
                raise TypeError(
                    "{} is a picklable extension while it's internal extension "
                    "{} is not.".format(type(self), type(extension))
                )
            if (not self.picklable and extension._runner == self.runner):
                # Cyclic referencing
                continue

            # Bind this extension to this binded extension.
            self.bind_new_extension(attr_name, extension, self.runner, self)

    def bind(self, runner, parent=None):
        """
        Binds the extension to the service runner and parent.
        :param Runner runner: Service Runner
        """
        if self.context_singleton and type(self) in runner.context_singletons:
            return runner.context_singletons[type(self)]

        if self.singleton and type(self) in runner.singletons:
            return runner.singletons[type(self)]

        if not self.picklable and self._runner == runner:
            return self

        ext = self._clone()
        if not self.picklable:
            ext.parent = parent
            ext._runner = runner

        ext.bind_inner_extensions()
        runner.register_extension(ext)
        return ext

    async def setup(self):
        """
        Called After the binding and before the system start
        """
        pass

    async def start(self):
        """
        Commands the extension to start running.
        """
        pass

    async def stop(self):
        """
        Commands the extension to do a graceful stop
        """
        pass

    def _clone(self):
        cls = type(self)
        return cls(*self._params[0], **self._params[1])

    @property
    def runner(self):
        if self._runner is None and not self.picklable:
            raise RuntimeError("This extension is not binded. You must bind "
                               "it to a runner before trying to attempt this.")
        return self._runner



import sys
import logging
import multiprocessing
import threading
from concurrent.futures import _base, ThreadPoolExecutor
from multiprocessing.queues import Queue
from queue import Empty

from micro_framework.exceptions import PoolStopped, BrokenSpawner
from micro_framework.spawners.base import Spawner, Task

logger = logging.getLogger(__name__)


def thread_worker(executor, write_queue, stop_event: threading.Event):
    while not stop_event.isSet():
        task_id = write_queue.get()
        with executor.shutdown_lock:
            if executor.shutdown_signal or task_id is None:
                # Shutdown signal
                return
        task = executor._tasks.pop(task_id, None)  # Remove to avoid Mem. inflating.
        if task is not None:
            task.run()


class ThreadSpawner(Spawner, _base.Executor):
    def __init__(self, config):
        self.max_workers = config.get("MAX_WORKERS")
        self._pool = []
        self._stop_map = {}
        self.ctx = multiprocessing.get_context()
        self.write_queue = Queue(ctx=self.ctx)
        self.shutdown_signal = False
        self.shutdown_lock = threading.Lock()
        self.initiate_pool()
        self.total_tasks = 0
        self._tasks = {}

    def _check_pool(self):
        """
        Verify the threads in the pool to spawn a new thread in case the
        count is different than the max_workers
        :return:
        """
        with self.shutdown_lock:
            if self.shutdown_signal:
                return
        logger.debug(f"Checking for dead threads in the pool. Currently "
                    f"{self._pool} threads in the pool.")

        for thread in self._pool:
            if not thread.is_alive():
                logger.debug(
                    f"Thread {thread} - PID: {thread.ident} is dead. Spawning "
                    f"a new thread to the pool."
                )
                self.stop_thread(thread)
                self.start_new_thread()

    def start_new_thread(self):
        evt = threading.Event()
        t = threading.Thread(
            target=thread_worker,
            args=(self, self.write_queue, evt)
        )
        t.daemon = True
        self._stop_map[t] = evt
        self._pool.append(t)
        t.start()
        return t

    def stop_thread(self, thread):
        self._pool.remove(thread)
        self._stop_map[thread].set()

    def initiate_pool(self):
        for _ in range(self.max_workers):
            self.start_new_thread()

    def create_task(self, target, *target_args, **target_kwargs):
        task = Task(self.total_tasks + 1, target, *target_args, **target_kwargs)
        return task

    def submit(self, target_fn, *args, **kwargs):
        with self.shutdown_lock:
            if self.shutdown_signal:
                raise PoolStopped("This pool is being shutdown.")
        task = self.create_task(target_fn, *args, **kwargs)
        self._tasks[task.task_id] = task
        self.write_queue.put(task.task_id)
        self.total_tasks += 1
        return task.future

    def shutdown(self, wait=True):
        with self.shutdown_lock:
            self.shutdown_signal = True
        # Threads don't have a terminate so in any case we would need to
        # gracefully stop it... The difference is that since it is a daemon,
        # we just skip joining the thread and leave to the parent process to
        # stop.
        logger.debug(f"Clearing Thread write queue "
                     f"({self.write_queue.qsize()})")
        while not self.write_queue.empty():  # Emptying the queue
            try:
                self.write_queue.get(timeout=1)
            except Empty:
                continue
        logger.debug("Sending stop signal to Threads")
        for thread in self._pool:
            self.write_queue.put(None)  # Signalling workers to stop
        if wait:
            logger.info("Gracefully shutting down thread workers...")
            for thread in self._pool:
                thread.join()  # Wait process to finish current task
        logger.debug("Thread Pool Finished")


class ThreadPoolSpawner(Spawner, ThreadPoolExecutor):
    def __init__(self, config):
        super(ThreadPoolSpawner, self).__init__(config.get("MAX_WORKERS"))

    def submit(self, *args, **kwargs):
        if sys.version_info.minor < 7:
            try:
                return super(ThreadPoolSpawner, self).submit(
                    *args, **kwargs
                )
            except BrokenPipeError as exc:
                raise BrokenSpawner from exc
        else:
            from concurrent.futures._base import BrokenExecutor
            try:
                return super(ThreadPoolSpawner, self).submit(*args, **kwargs)
            except (BrokenExecutor, BrokenPipeError) as exc:
                raise BrokenSpawner from exc

    def shutdown(self, **kwargs) -> None:
        logger.info("Greedy Worker shutdown initiated, wait until all pending "
                    "tasks are consumed.")
        super(ThreadPoolSpawner, self).shutdown(**kwargs)



import logging
from threading import Lock

from kombu import Consumer
from kombu.mixins import ConsumerMixin

from micro_framework.amqp.amqp_elements import rpc_queue, rpc_exchange, \
    rpc_broadcast_queue, get_connection, Publisher, \
    get_rpc_target_id_from_routing_key
from micro_framework.amqp.exceptions import DuplicatedListener
from micro_framework.exceptions import ExtensionIsStopped, PoolStopped, \
    RPCTargetDoesNotExist
from micro_framework.extensions import Extension
from micro_framework.rpc import RPCManagerMixin, format_rpc_response

logger = logging.getLogger(__name__)


class ConsumerManager(Extension, ConsumerMixin):
    def __init__(self):
        self._clean()
        self.queues = {}
        self.run_thread = None
        self.connection = None
        # Lock due to message object apparently not being thread-safe...
        self.message_lock = Lock()
        self.started = False

    context_singleton = True

    @property
    def amqp_uri(self):
        return self.runner.config['AMQP_URI']

    def _clean(self):
        self._consumers = []
        self.entrypoint_map = {}  # exchange.routing_key -> Entrypoint

    def on_connection_error(self, exc, interval):
        logger.info(
            "Error connecting to broker at {} ({}).\n"
            "Retrying in {} seconds.".format(self.amqp_uri, exc, interval)
        )
        logger.debug(
            "Removing all previous consumers, since they will be "
            "loaded again."
        )
        self._clean()

    def _entrypoint_key(self, exchange, routing_key):
        return f"{exchange}.{routing_key}"

    async def get_connection(self):
        if not self.connection:
            self.connection = get_connection(self.amqp_uri)
        return self.connection

    async def setup(self):
        self.connection = await self.get_connection()

    async def start(self):
        if not self.started:
            self.started = True
            await self.runner.spawn_extension(self, self.run)

    async def stop(self):
        if self.started:
            logger.debug("AMQP ConsumerManager is stopping")
            self.should_stop = True
            logger.debug("AMQP ConsumerManager stopped.")
            self.started = False

    async def add_entrypoint(self, entrypoint):
        queue = entrypoint.queue
        self.queues[queue] = entrypoint

    def get_consumers(self, _, channel):
        internal_queues = set()
        normal_queues = set()

        for queue, entrypoint in self.queues.items():


            if entrypoint.internal:
                internal_queues.add(queue)
            else:
                normal_queues.add(queue)

            entrypoint_key = self._entrypoint_key(
                queue.exchange.name, queue.routing_key
            )
            if entrypoint_key in self.entrypoint_map:
                raise DuplicatedListener(
                    "You cannot define a listener to the same set "
                    "({}, {}) in the same RunnerContext.".format(
                        queue.exchange.name, queue.routing_key
                    )
                )
            self.entrypoint_map[entrypoint_key] = entrypoint

        self._consumers = [
            Consumer(
                channel,
                queues=normal_queues,
                callbacks=[self.on_message],
                no_ack=False,
                prefetch_count=self.runner.max_workers
            ),
            Consumer(
                channel.connection.channel(),
                queues=internal_queues,
                callbacks=[self.on_message],
                no_ack=False,
            )
        ]
        return self._consumers

    def on_consume_ready(self, connection, channel, consumers, **kwargs):
        logger.debug(f'consumer started {consumers}')
        self.connection.connect()

    async def handle_new_message(self, entrypoint, body, message):
        try:
            await entrypoint.call_route(message, body)
        except Exception:
            logger.exception("")
            await self.requeue_message(message)
        else:
            await self.ack_message(message)

    def on_message(self, body, message):
        logger.debug("Message Received")
        entrypoint_key = self._entrypoint_key(
            message.delivery_info["exchange"],
            message.delivery_info["routing_key"]
        )
        entrypoint = self.entrypoint_map.get(entrypoint_key)

        if entrypoint is None:
            message.ack()
            return

        if not self.should_stop:
            try:
                self.runner.execute_async(
                    self.handle_new_message(entrypoint, body, message),
                    raise_for_future=False
                )
            except (PoolStopped, ExtensionIsStopped):
                message.requeue()
        else:
            message.requeue()

    async def ack_message(self, message):
        message.ack()

    async def requeue_message(self, message):
        message.requeue()


class RPCManager(RPCManagerMixin, ConsumerMixin):
    def __init__(self):
        super(RPCManager, self).__init__()
        self.connection = None
        self.started = False
        self.message_lock = Lock()

    context_singleton = True

    @property
    def amqp_uri(self):
        return self.runner.config['AMQP_URI']

    async def get_connection(self):
        if not self.connection:
            self.connection = get_connection(self.amqp_uri)
        return self.connection

    async def setup(self):
        self.connection = await self.get_connection()
        self.publisher = Publisher(self.amqp_uri)

    async def start(self):
        if not self.started:
            self.started = True
            await self.runner.spawn_extension(self, self.run)

    async def stop(self):
        if self.started:
            logger.debug("AMQP RPCManager is stopping")
            self.should_stop = True
            self.connection.close()
            logger.debug("AMQP RPCManager stopped.")
            self.started = False

    def get_consumers(self, _, channel):
        exchange = rpc_exchange()
        queues = set()
        broadcast_queue = rpc_broadcast_queue(
            self.runner.config["SERVICE_NAME"]
        )

        for target_id, entrypoint in self.entrypoints.items():
            queue = rpc_queue(
                self.runner.config["SERVICE_NAME"], target_id, exchange
            )
            queues.add(queue)

        self._consumers = [
            Consumer(
                channel,
                queues=queues,
                callbacks=[self.on_message],
                no_ack=False,
                prefetch_count=self.runner.max_workers
            ),
            Consumer(
                channel.connection.channel(), # Get new channel.
                queues=[broadcast_queue],
                callbacks=[self.on_broadcast_message],
                no_ack=False,
            )
        ]

        return self._consumers

    async def send_reply(self, message, payload):
        """
        Send the RPC Response to the Reply Queue
        :param message:
        :param future:
        :return:
        """
        exchange = rpc_exchange()
        routing_key = message.properties.get("reply_to")
        correlation_id = message.properties.get("correlation_id")
        if not routing_key or not correlation_id:
            # Nothing to do here. Might be some unknown message.
            return

        await self.runner.sync_to_async(
            self.publisher.publish, payload, exchange=exchange,
            routing_key=routing_key, correlation_id=correlation_id, retry=True
        )

    async def handle_new_call(self, target_ids, body, message):
        """
        Async Method to handle all procedures when receiving a RPC call.
        """
        try:
            # We verify if this callback is the one that handles the given
            # target_ids
            routing_key = message.delivery_info["routing_key"]
            target_id = routing_key.split(".", 1)[1]
            if "_broadcast." in target_id:
                target_id = target_id.split("_broadcast.", 1)[1]

            if target_id in target_ids:
                response = await self.consume_message(message, body)
            else:
                response = format_rpc_response(
                    None, exception=RPCTargetDoesNotExist(
                        "This target does not exist in the called RPC method"
                    )
                )
            await self.send_reply(message, response)
        except Exception:
            logger.exception("Error on handling a RPCManager call")
            await self.requeue_message(message)
        else:
            await self.ack_message(message)

    def on_message(self, body, message):
        logger.debug("AMQP RPC Message Received")
        if not self.should_stop:
            try:

                target_id = get_rpc_target_id_from_routing_key(
                    self.runner.config["SERVICE_NAME"],
                    message.delivery_info["routing_key"]
                ) # We can infer the target by getting the routing key.

                # Calling the async method to handle the new message.
                self.runner.execute_async(
                    self.handle_new_call([target_id], body, message),
                    raise_for_future=False
                )
                return
            except (ExtensionIsStopped, PoolStopped):
                message.requeue()
        else:
            message.requeue()

    def on_broadcast_message(self, body, message):
        logger.debug("AMQP RPC Broadcast Message Received")
        if not self.should_stop:
            try:
                # Calling the async method to handle the new message.
                self.runner.execute_async(
                    self.handle_new_call(self.internal_commands, body, message)
                )
                return
            except (ExtensionIsStopped, PoolStopped):
                message.requeue()
        else:
            message.requeue()

    async def ack_message(self, message):
        message.ack()

    async def requeue_message(self, message):
        message.requeue()



import logging
from wsgiref.simple_server import make_server

from prometheus_client import make_wsgi_app, REGISTRY, Gauge, Counter, Summary, \
    Info
from prometheus_client.exposition import ThreadingWSGIServer, _SilentHandler

from micro_framework.extensions import Extension

logger = logging.getLogger(__name__)


class PrometheusMetricServer(Extension):
    singleton = True

    def __init__(self):
        self.started = False

    async def setup(self):
        self.config = self.runner.config['METRICS']
        self.enabled = self.runner.config['ENABLE_METRICS']
        self.started = False

    async def start(self):
        """
        Starts a WSGI server exactly like prometheus_client.start_http_server
        except that we use our runner to spawn the task instead of a Thread
        alone.
        """
        # TODO Change to uvicorn
        if self.enabled:
            app = make_wsgi_app(REGISTRY)
            self.httpd = make_server(
                self.config['HOST'],
                self.config['PORT'],
                app,
                ThreadingWSGIServer,
                handler_class=_SilentHandler
            )
            logger.info(
                "Prometheus Metrics Server Started\n\t-Listening at"
                f" {self.config['HOST']}:{self.config['PORT']}"
            )
            self.started = True
            self.task = self.runner.event_loop.run_in_executor(
                None, self.httpd.serve_forever
            )

    async def stop(self):
        if self.started:
            self.httpd.shutdown()


# Runner Metrics
tasks_number = Gauge(
    'tasks_number', 'Number of Tasks being run at the moment',
    ["context", 'route']
)

tasks_counter = Counter(
    'tasks_number', 'Total number of tasks',
    ["context", "route"]
)

tasks_failures_counter = Counter(
    'failed_tasks', 'Total number of tasks that have failed',
    ["context", "route"]
)

tasks_duration = Summary(
    'task_duration_seconds', 'Duration time of a task',
    ["context", "route"]
)

info = Info(
    'service', 'Information regarding the service', ["context"]
)

available_workers = Gauge(
    "available_workers", "Number of Workers available",
    ["context"]
)



import signal
from concurrent.futures import _base
from logging import getLogger

from loky.backend import get_context

logger = getLogger(__name__)


class Spawner:
    """
    Mixin to Executors
    """

    def spawn(self, target_fn, *fn_args, callback=None, **fn_kwargs):
        fn_args = fn_args or tuple()
        fn_kwargs = fn_kwargs or dict()
        logger.debug(f"Spawning {target_fn.__name__}")
        future = self.submit(target_fn, *fn_args, **fn_kwargs)
        logger.debug(f"Target {target_fn.__name__} spawned")
        if callback:
            future.add_done_callback(callback)
        return future

    def stop(self, wait=True):
        self.shutdown(wait=wait)

    def kill(self):
        self.shutdown(wait=False)


def worker_init():
    signal.signal(signal.SIGINT, signal.SIG_IGN)


class CallItem:
    def __init__(self, task_id, target, *target_args, **target_kwargs):
        self.task_id = task_id
        self.target = target
        self.target_args = target_args
        self.target_kwargs = target_kwargs

    @classmethod
    def from_task(cls, task):
        return cls(task.task_id, task.target, *task.target_args,
                   **task.target_kwargs)


class Task:
    def __init__(self, task_id, target, *target_args, **target_kwargs):
        self.task_id = task_id
        self.target = target
        self.target_args = target_args
        self.target_kwargs = target_kwargs
        self.future = _base.Future()

    def run(self):
        # Called by Thread Pool
        try:
            result = self.target(*self.target_args, **self.target_kwargs)
            self.future.set_result(result)
        except Exception as _exc:
            self.future.set_exception(_exc)


ctx = get_context()


import asyncio
import inspect
import logging
from threading import Thread

from functools import partial

logger = logging.getLogger(__name__)


def _start_worker_event_loop():
    """
    Starts an event-loop and set is as the event-loop for the current process.

    The event-loop would be already running inside a Thread, returned from
    the function.

    It should be called only for process workers that spawn a new process
    and therefore can't access the main event-loop
    """
    event_loop = asyncio.new_event_loop()
    asyncio.set_event_loop(event_loop)
    event_loop_thread = Thread(
        target=event_loop.run_forever, daemon=True
    )
    event_loop_thread.start()

    return event_loop, event_loop_thread


def executor_task(func, *fn_args, **fn_kwargs):
    """
    Function to execute a task using a spawner.

    :param func: Function to be called
    :param fn_args: Function args
    :param fn_kwargs: Function kwargs
    :return: Function Result
    :raises: Function Exceptions
    """
    # TODO This fn might be responsibility of target module
    event_loop_thread = None
    event_loop = asyncio.new_event_loop()

    try:
        if inspect.iscoroutinefunction(func):
            # run the coroutine using a new event loop
            asyncio.set_event_loop(event_loop)
            result = event_loop.run_until_complete(
                func(*fn_args, **fn_kwargs)
            )

        else:
            # Run the function but starts a running event loop before,
            # in order to the Target methods that are asyncio.
            # TODO this is why this fn seems like it belongs to targets.py
            event_loop, event_loop_thread = _start_worker_event_loop()
            result = func(*fn_args, **fn_kwargs)
    finally:
        event_loop.call_soon_threadsafe(event_loop.stop)

        if event_loop_thread:
            event_loop_thread.join()  # wait until completely closed.
        event_loop.close()
    return result


class Worker:
    """
    Worker class that calls the target using an executor.

    The Task is a TargetFunction instance or a TargetClass instance.

    The worker is responsible to call the Target methods:
        * on_worker_setup
        * mount_target
        * on_worker_finished
    """

    def __init__(
            self, target, translators, config, *fn_args, _meta=None,
            **fn_kwargs
    ):
        self.target = target
        self.config = config
        self.fn_args = fn_args
        self.fn_kwargs = fn_kwargs
        self._translators = translators or []
        self.result = None
        self.exception = None
        self.finished = False
        self._meta = _meta or {}  # Content shared by extensions

    async def call_task(self, spawner, mounted_target, *fn_args, **fn_kwargs):
        try:

            if spawner: # Run it outside the event-loop
                event_loop = asyncio.get_event_loop()
                fn = partial(
                    executor_task, mounted_target.run, *fn_args, **fn_kwargs
                )

                self.result = await event_loop.run_in_executor(spawner, fn)

            else:  # When no spawner is given, we consider it an asyncio target.
                self.result = await mounted_target.run(
                    *fn_args, **fn_kwargs
                )

        except Exception as exc:
            self.exception = exc
            logger.exception("")

        return self.result

    async def run(self, runner):
        self.runner = runner
        await self.target.on_worker_setup(self)

        # Translating Messages if there is any translator
        self.translated_args = self.fn_args
        self.translated_kwargs = self.fn_kwargs
        for translator in self._translators:
            self.translated_args, self.translated_kwargs = await translator.translate(
                *self.fn_args, **self.fn_kwargs
            )

        mounted_target = await self.target.mount_target(self)
        await self.call_task(
            runner.spawner, mounted_target, *self.translated_args,
            **self.translated_kwargs
        )
        self.finished = True
        await self.target.on_worker_finished(self)
        return self


class CallbackWorker(Worker):
    def __init__(self, callback_target, original_worker):
        self.callback_target = callback_target
        self.original_worker = original_worker
        args = (*original_worker.translated_args, original_worker.exception)
        kwargs = original_worker.translated_kwargs
        # We use the already translated content but the dependencies will
        # have to be re-created due to the pickling problem of Processing.
        super(CallbackWorker, self).__init__(
            callback_target, None,
            original_worker.config, *args, _meta=original_worker._meta,
            **kwargs
        )



from .multiprocess import ProcessPoolSpawner
from .thread import ThreadSpawner, ThreadPoolSpawner

# Greedy Spawner means that the runner will consume all received tasks so far
# when a stop signal is sent.

# The non-greedy spawner will finish the current tasks and skip all pending
# ones. (Entrypoints with acknowledge are good candidates for this)

SPAWNERS = {
    'thread': ThreadSpawner,
    'process': ProcessPoolSpawner,
    'greedy_thread': ThreadPoolSpawner
}


THREAD_SPAWNERS = [ThreadSpawner, ThreadPoolSpawner]

PROCESS_SPAWNERS = [ProcessPoolSpawner]



from contextlib import contextmanager

from abc import ABC, abstractmethod


class RPCConnection(ABC):
    """
    Abstract Class representing an instantiated connection.
    """

    @abstractmethod
    def send(self, *args, **kwargs):
        """
        Send a message without waiting for an answer.
        """
        ...

    @abstractmethod
    def send_and_receive(self, *args, **kwargs):
        """
        Send a message and wait for the connected server response.
        """
        ...


class RPCConnector(ABC):
    """
    Class responsible to provide an interface to obtain a new connection.

    Each time the get_connection method is called, it should provide a new
    connection instance yielding it.

    All connection handling should be done inside the get_connection method,
    since this might not be thread-safe.
    """

    @abstractmethod
    @contextmanager
    def get_connection(self) -> RPCConnection:
        """
        Returns an open connection that implements RPCConnection.
        """
        ...



from logging import getLogger
from typing import Any

from kombu import Exchange, Queue

from micro_framework.amqp.amqp_elements import event_queue_name, \
    event_exchange_name
from micro_framework.amqp.manager import ConsumerManager, RPCManager
from micro_framework.entrypoints import Entrypoint
from micro_framework.rpc.manager import target_detail

logger = getLogger(__name__)


class BaseEventListener(Entrypoint):
    """
    Base class to handle Queue consuming of a single routing key in an exchange.
    It declares a Queue, binded to the exchange with exchange_name and with a
    routing_key from routing_key attribute.
    """
    def __init__(self, exchange_name=None, routing_key=None,
                 exchange_type='topic', payload_filter=None):
        self.exchange_name = exchange_name
        self.routing_key = routing_key
        self.exchange_type = exchange_type
        self.payload_filter = payload_filter

    manager = ConsumerManager()
    internal = False

    async def get_exchange(self):
        """
        Return the Exchange that will have a queue binded to.
        :return Exchange:
        """
        return Exchange(
            name=self.exchange_name, type=self.exchange_type,
            auto_delete=True
        )

    async def get_queue_name(self):
        """
        Return the name of the queue to be binded.
        It usually must be unique by consuming entrypoint to avoid the same
        queue being consumed by multiple different routes.
        :return str: The name of the Queue
        """
        return f'{self.exchange_name}_{self.routing_key}_queue'

    async def get_queue(self):
        """
        Returns the Queue to be created or loaded using the exchange and
        routing key declared.
        :return:
        """
        queue_name = await self.get_queue_name()
        exchange = await self.get_exchange()
        return Queue(
            name=queue_name, exchange=exchange,
            routing_key=self.routing_key, durable=True
        )

    async def setup(self):
        await super(BaseEventListener, self).setup()
        self.exchange = await self.get_exchange()
        self.queue = await self.get_queue()
        self.queue.maybe_bind(await self.manager.get_connection())
        self.queue.declare()
        await self.manager.add_entrypoint(self)

    async def call_route(self, entry_id, *args, _meta=None, **kwargs):
        """
        Called by the Manager when a new message is received.
        :param Message message: Message object
        :param dict|str payload: Event Payload
        """
        payload = args[0]
        if not self.payload_filter or self.payload_filter(payload):
            return await super(BaseEventListener, self).call_route(
                entry_id, *args, _meta=_meta, **kwargs
            )


class QueueListener(BaseEventListener):
    def __init__(self, *args, queue_name=None, **kwargs):
        super(QueueListener, self).__init__(*args, **kwargs)
        self.queue_name = queue_name

    async def get_queue_name(self):
        if self.queue_name:
            return self.queue_name
        return await super(QueueListener, self).get_queue_name()


class EventListener(QueueListener):
    def __init__(self, source_service, event_name, **kwargs):
        self.source_service = source_service
        self.event_name = event_name
        exchange_name = event_exchange_name(source_service)
        super(EventListener, self).__init__(exchange_name, event_name, **kwargs)

    async def get_queue_name(self):
        service_name = self.runner.config['SERVICE_NAME']
        return event_queue_name(
            self.event_name, self.source_service,
            service_name, self.route.target
        )


class RPCListener(Entrypoint):
    manager = RPCManager()

    async def setup(self):
        await super(RPCListener, self).setup()
        await self.manager.add_entrypoint(self)



import uuid
from contextlib import contextmanager

import warnings
from amqp import ChannelError
from kombu import Queue, Exchange, Connection, producers

from micro_framework.exceptions import RPCTargetDoesNotExist
from micro_framework.rpc import RPCConnection
from micro_framework.targets import Target

rpc_exchange_name = "ufw-rpc"
DEFAULT_TRANSPORT_OPTIONS = {
    'max_retries': 3,
    'interval_start': 2,
    'interval_step': 1,
    'interval_max': 5
}
DEFAULT_RETRY_POLICY = {'max_retries': 3}
DEFAULT_HEARTBEAT = 60
DEFAULT_SERIALIZER = 'json'
PERSISTENT = 2


def event_queue_name(event_name: str, source_service: str, service_name: str,
                     target: Target) -> str:
    """
    Returns the Queue name that should be used when listening to events from
    a publisher.

    :param str event_name:
    :param str source_service:
    :param str service_name:
    :param Target target:
    :return str: The Queue name
    """

    # source_service.something_happened__my_service.target.function
    return f'{source_service}.{event_name}__{service_name}.{target}'


def event_exchange_name(source_service: str) -> str:
    return f"{source_service}.events"


def rpc_queue_name(service_name, target_id):
    return "rpc-{}-{}".format(service_name, target_id)


def rpc_routing_key(service_name, target_id):
    return "{}.{}".format(service_name, target_id)


def get_rpc_target_id_from_routing_key(service_name, routing_key):
    _, target_id = routing_key.split("{}.".format(service_name))
    return target_id


def rpc_broadcast_routing_key(service_name, target_id):
    return "{}._broadcast.{}".format(service_name, target_id)


def rpc_broadcast_queue(service_name):
    return Queue(
        "rpc-{}-broadcast".format(service_name),
        exchange=rpc_exchange(),
        routing_key="{}._broadcast.*".format(service_name)
    )


def rpc_queue(service_name: str, target_id: str, exchange: Exchange) -> Queue:
    return Queue(
        rpc_queue_name(service_name, target_id),
        exchange=exchange,
        routing_key=rpc_routing_key(service_name, target_id)
    )


def rpc_exchange() -> Exchange:
    return Exchange(name=rpc_exchange_name, durable=True, type="topic")


def rpc_reply_queue() -> Queue:
    """
    Returns a Queue instance for RPC Replies.

    This Queue is Durable and each message will live for a day without being
    consumed.

    Since at each start we don't want to route messages to the wrong service,
    the routing key should be unique.

    :return Queue: The RPC Reply Queue.
    """
    exchange = rpc_exchange()
    name = f"ufw-reply-{uuid.uuid4()}"
    return Queue(
        name, exchange=exchange,
        routing_key=name, message_ttl=3600 * 24,
        expires=3600 * 12 # Stays 12 Hours before being deleted.
    )


default_transport_options = {
    'confirm_publish': True
}


def get_connection(amqp_uri):
    connection = Connection(
        amqp_uri, transport_options=default_transport_options,
        heartbeat=120
    )
    return connection


@contextmanager
def get_producer(amqp_uri, confirms=True, ssl=None, transport_options=None):
    if transport_options is None:
        transport_options = DEFAULT_TRANSPORT_OPTIONS.copy()
    transport_options['confirm_publish'] = confirms
    conn = Connection(amqp_uri, transport_options=transport_options, ssl=ssl)

    with producers[conn].acquire(block=True) as producer:
        yield producer


class Publisher:
    """
    Utility helper for publishing messages to RabbitMQ.

    ################### IMPORTANT NOTE #############################
    #                                                              #
    #   THIS CLASS WAS (NOT PROUDLY) TAKEN FROM NAMEKO PROJECT     #
    #                                                              #
    #    Don't know why, but my implementation wasn't              #
    #    working properly                                          #
    #                                                              #
    ################################################################

    ref: https://github.com/nameko/nameko/blob/a719cb1487f643769e2d13daf255c20551490f43/nameko/amqp/publish.py#L40
    """

    use_confirms = True
    """
    Enable `confirms <http://www.rabbitmq.com/confirms.html>`_ for this
    publisher.
    The publisher will wait for an acknowledgement from the broker that
    the message was receieved and processed appropriately, and otherwise
    raise. Confirms have a performance penalty but guarantee that messages
    aren't lost, for example due to stale connections.
    """

    transport_options = DEFAULT_TRANSPORT_OPTIONS.copy()
    """
    A dict of additional connection arguments to pass to alternate kombu
    channel implementations. Consult the transport documentation for
    available options.
    """

    delivery_mode = PERSISTENT
    """
    Default delivery mode for messages published by this Publisher.
    """

    mandatory = False
    """
    Require `mandatory <https://www.rabbitmq.com/amqp-0-9-1-reference.html
    #basic.publish.mandatory>`_ delivery for published messages.
    """

    priority = 0
    """
    Priority value for published messages, to be used in conjunction with
    `consumer priorities <https://www.rabbitmq.com/priority.html>_`.
    """

    expiration = None
    """
    `Per-message TTL <https://www.rabbitmq.com/ttl.html>`_, in milliseconds.
    """

    serializer = "json"
    """ Name of the serializer to use when publishing messages.
    Must be registered as a
    `kombu serializer <http://bit.do/kombu_serialization>`_.
    """

    compression = None
    """ Name of the compression to use when publishing messages.
    Must be registered as a
    `kombu compression utility <http://bit.do/kombu-compression>`_.
    """

    retry = True
    """
    Enable automatic retries when publishing a message that fails due
    to a connection error.
    Retries according to :attr:`self.retry_policy`.
    """

    retry_policy = DEFAULT_RETRY_POLICY
    """
    Policy to apply when retrying message publishes, if requested.
    See :attr:`self.retry`.
    """

    declare = []
    """
    Kombu :class:`~kombu.messaging.Queue` or :class:`~kombu.messaging.Exchange`
    objects to (re)declare before publishing a message.
    """

    def __init__(
            self, amqp_uri, use_confirms=None, serializer=None,
            compression=None,
            delivery_mode=None, mandatory=None, priority=None, expiration=None,
            declare=None, retry=None, retry_policy=None, ssl=None,
            **publish_kwargs
    ):
        self.amqp_uri = amqp_uri
        self.ssl = ssl

        # publish confirms
        if use_confirms is not None:
            self.use_confirms = use_confirms

        # delivery options
        if delivery_mode is not None:
            self.delivery_mode = delivery_mode
        if mandatory is not None:
            self.mandatory = mandatory
        if priority is not None:
            self.priority = priority
        if expiration is not None:
            self.expiration = expiration

        # message options
        if serializer is not None:
            self.serializer = serializer
        if compression is not None:
            self.compression = compression

        # retry policy
        if retry is not None:
            self.retry = retry
        if retry_policy is not None:
            self.retry_policy = retry_policy

        # declarations
        if declare is not None:
            self.declare = declare

        # other publish arguments
        self.publish_kwargs = publish_kwargs

    def publish(self, payload, **kwargs):
        """ Publish a message.
        """
        publish_kwargs = self.publish_kwargs.copy()

        # merge headers from when the publisher was instantiated
        # with any provided now; "extra" headers always win
        headers = publish_kwargs.pop('headers', {}).copy()
        headers.update(kwargs.pop('headers', {}))
        headers.update(kwargs.pop('extra_headers', {}))

        use_confirms = kwargs.pop('use_confirms', self.use_confirms)
        transport_options = kwargs.pop('transport_options',
                                       self.transport_options
                                       )
        transport_options['confirm_publish'] = use_confirms

        delivery_mode = kwargs.pop('delivery_mode', self.delivery_mode)
        mandatory = kwargs.pop('mandatory', self.mandatory)
        priority = kwargs.pop('priority', self.priority)
        expiration = kwargs.pop('expiration', self.expiration)
        serializer = kwargs.pop('serializer', self.serializer)
        compression = kwargs.pop('compression', self.compression)
        retry = kwargs.pop('retry', self.retry)
        retry_policy = kwargs.pop('retry_policy', self.retry_policy)

        declare = self.declare[:]
        declare.extend(kwargs.pop('declare', ()))

        publish_kwargs.update(kwargs)  # remaining publish-time kwargs win

        with get_producer(self.amqp_uri,
                          use_confirms,
                          self.ssl,
                          transport_options,
                          ) as producer:
            try:
                producer.publish(
                    payload,
                    headers=headers,
                    delivery_mode=delivery_mode,
                    mandatory=mandatory,
                    priority=priority,
                    expiration=expiration,
                    compression=compression,
                    declare=declare,
                    retry=retry,
                    retry_policy=retry_policy,
                    serializer=serializer,
                    **publish_kwargs
                )
            except ChannelError as exc:
                if "NO_ROUTE" in str(exc):
                    raise RPCTargetDoesNotExist()
                raise

            if mandatory and not use_confirms:
                warnings.warn(
                    "Mandatory delivery was requested, but "
                    "unroutable messages cannot be detected without "
                    "publish confirms enabled."
                )



from micro_framework.entrypoints import Entrypoint
from micro_framework.exceptions import MaxConnectionsReached
from micro_framework.websocket.manager import WebSocketManager


class WebSocketEntrypoint(Entrypoint):
    manager = WebSocketManager()

    async def call_route(self, entry_id, *args, _meta=None, **kwargs):
        if self.runner.available_workers <= 0:
            raise MaxConnectionsReached("No Available Workers")
        return await super(WebSocketEntrypoint, self).call_route(
            entry_id, *args, _meta=_meta, **kwargs
        )

    async def setup(self):
        await super(WebSocketEntrypoint, self).setup()
        await self.manager.add_entrypoint(self)



from contextlib import contextmanager

import websockets
from websocket import create_connection, WebSocket

from micro_framework.exceptions import FrameworkException
from micro_framework.rpc import RPCConnection
from micro_framework.rpc.connectors import RPCConnector


class ServerDisconnected(FrameworkException):
    pass


class WSConnection(RPCConnection):
    def __init__(self, uri, timeout):
        self.uri = uri
        self.timeout = timeout

    @contextmanager
    def connection(self):
        con = create_connection(
            self.uri, timeout=self.timeout, enable_multithread=True
        )
        yield con
        con.close()

    def send(self, *args, **kwargs):
        with self.connection() as con:
            con.send(*args, **kwargs)

    def send_and_receive(self, payload, *args, **kwargs):
        with self.connection() as con:
            con.send(payload, *args, **kwargs)
            return con.recv()


class WebSocketConnector(RPCConnector):
    def __init__(self, uri, timeout=None):
        self.uri = uri
        self.timeout = timeout

    def get_connection(self) -> RPCConnection:
        return WSConnection(self.uri, self.timeout)



import inspect
import sys

from functools import wraps
from typing import Any, Callable

from micro_framework.extensions import Extension

from micro_framework.workers import Worker

def inject(**dependencies):
    for dep in dependencies.values():
        if not isinstance(dep, Dependency):
            raise TypeError(
                f"Injected Dependency {type(dep)} should be a subclass of "
                f"{Dependency}"
            )
    def decorator(fn):
        if inspect.iscoroutinefunction(fn):
            @wraps(fn)
            async def wrapped(*args, **kwargs):
                return await fn(*args, **kwargs)
        else:
            @wraps(fn)
            def wrapped(*args, **kwargs):
                return fn(*args, **kwargs)
        setattr(sys.modules[fn.__module__], fn.__name__, wrapped)
        wrapped.__dependencies__ = dependencies
        return wrapped
    return decorator


class Dependency(Extension):
    """
    Kind of extension that injects a dependency into the worker target.

    The Dependency instance will be used by multiple workers so it is not
    recommended to store states to be used between the methods calls unless
    it is desirable to share this states between all workers of the same
    Route.

    If you wish to store some values to be used between the methods in the
    same worker, use the worker instance passed.

    The Dependency Life-Cycle is:

    ###### Main Process/Thread #########

    1. setup:
        . Like any Extension, it has a setup method. Usually not used for
        this kind of Dependency.

    2. Setup/Get RunnerDependencies and inject them in the worker call

    ###### Worker Process/Thread ########

    3. setup_dependency:
        .Do anything that might be needed before actually creating the
        injected dependency.

    4. get_dependency:
        .Returns the dependency to be injected

    5. Run the Target Function/Method passing the injected dependency

    6. after_call:
        .After the worker is executed, we call this method passing the
        worker and the result + exception from that call.

    ###### Main Process/Thread #########

    7. after_call of RunnerDependency classes.

    """
    picklable = True

    def bind(self, runner, parent=None):
        ext = super(Dependency, self).bind(runner, parent=None)
        ext.config = runner.config
        return ext

    async def setup_dependency(self, worker: Worker):
        """
        Do any setup necessary before the dependency injection.
        """
        pass

    async def get_dependency(self, worker: Worker) -> Callable:
        """
        Injects the dependency that will be passed to the Target Function.
        """
        pass

    async def after_call(self, worker: Worker, result: Any, exc: Exception):
        """
        Cleanup after the function has finished or raised.
        """
        pass


class RunnerDependency(Dependency):
    """
    A type of Dependency that is loaded before sending the task to the
    executor.

    This Dependency has access to the runner instance and therefore can spawn
    extensions or run asyncio coroutines inside the runner event-loop.

    The user should avoid this type of Dependency when working with Processes,
    since it can cause some pickling problems, due to the injected dependency
    being sent before the spawning of the worker instead of after like the
    Dependency class.

    The RunnerDependency Life-Cycle is:

    ###### Main Process/Thread #########

    1. setup:
        . Since this kind of dependency has access to the runner, we might
        setup some kind of internal worker here.

    2. start:
        . Start the RunnerDependency internals

    3. setup_dependency:
        .Do anything that might be needed before actually creating the
        injected dependency.

    4. get_dependency:
        .Returns the dependency to be injected

    ###### Worker Process/Thread ########

    5. setup/get_dependency of normal Dependency classes

    6. Run the Target Function/Method with injected dependencies

    7. after_call of normal Dependency classes

    ###### Main Process/Thread #########

    8. after_call:
        .After the worker is executed, we call this method passing the
        worker and the result + exception from that call.

    """
    picklable = False

    def bind(self, runner, parent=None):
        ext = super(Dependency, self).bind(runner, parent=parent)
        ext.target = parent
        ext.config = runner.config
        return ext

    async def setup_dependency(self, worker: Worker):
        """
        Do any setup necessary before the dependency injection.
        """
        pass

    async def get_dependency(self, worker: Worker) -> Callable:
        """
        Injects the dependency that will be passed to the Target Function.

        * NOTE:
            Remember that this will be called before spawning a worker
            and there are restrictions.

            Currently we use loky as a process spawner and it uses
            cloudpickle, so search about it's limitations.
        """
        pass

    async def after_call(self, worker: Worker, result: Any, exc: Exception):
        """
        Cleanup after the function has finished or raised.
        """
        pass


class Config(Dependency):
    """
    Returns the Runner Configuration.
    """
    async def get_dependency(self, worker: Worker):
        return self.config



import asyncio
import logging
from datetime import datetime

from kombu import Exchange, Queue

from micro_framework.amqp.connectors import PublisherClient
from micro_framework.amqp.dependencies import dispatch
from micro_framework.amqp.entrypoints import BaseEventListener
from micro_framework.dependencies import Dependency
from micro_framework.extensions import Extension
from micro_framework.routes import Route

logger = logging.getLogger(__name__)


class BackOffContext:
    """
    Object to be passed as argument to functions that implement a BackOff.
    """

    def __init__(self, retry_count, last_retry):
        self.retry_count = retry_count
        self.last_retry = last_retry


class BackOffData(Dependency):
    """
    Dependency Provider that loads the current backoff metadata from the
    worker and passes it to the user as a BackOffContext object.
    """

    def get_dependency(self, worker):
        backoff_meta = worker._meta.get('backoff', {})
        return BackOffContext(
            retry_count=backoff_meta.get('retry_count', 0),
            last_retry=backoff_meta.get('last_retry', None)
        )


class BackOff(Extension):
    """
    Adds BackOff Capability to a failed route. It is responsible to firing
    retries when needed.
    """

    def __init__(self, max_retries=3, interval=60000, exception_list=None):
        """
        :param int max_retries: Total number of retry attempts.
        :param interval: Interval between retries in micro seconds
        :param exception_list: list of exceptions that enable retries.
        """
        self.max_retries = max_retries
        self.interval = interval  # Microseconds
        self.exception_list = exception_list
        self.route = None

    async def get_interval(self, context):
        """
        Return the interval to be used on the next retry action.
        If you wish to implement custom intervals, this is the method to
        override.

        The context received is the context before updating so the
        retry_count starts with zero and the last_retry is None at first.

        :param dict context: BackOff meta data
        :return int: Interval in milliseconds.
        """
        return self.interval

    async def get_max_retries(self, context):
        """
        Return the total of retries to be compared with the current number of
        retries in order to decide to allow retry or not.
        :param dict context: BackOff meta data.
        :return int: Maximum number of retries.
        """
        return self.max_retries

    async def get_exception_list(self, context):
        """
        Returns the list of exceptions to be compared with the current worker
        exception in order to allow the retry action.

        If None is returned, then we allow all exceptions.
        :param dict context: BackOff meta data.
        :return list/None : List of Exceptions or None to allow all.
        """
        return self.exception_list

    async def _exception_in_list(self, exception, context):
        exception_list = await self.get_exception_list(context)
        if not exception_list:
            return True
        return type(exception) in exception_list

    async def _retry_count_is_lower(self, context):
        retry_count = context['retry_count']
        if retry_count < await self.get_max_retries(context):
            return True
        return False

    async def _interval_passed(self, context):
        last_retry = context['last_retry'] or 0
        if last_retry + await self.get_interval(
                context) / 1000 < datetime.utcnow().timestamp():
            return True
        return False

    async def _get_context(self, worker):
        context = worker._meta.get('backoff', None)
        if context is None:
            context = {
                'retry_count': 0, 'last_retry': None
            }
        return context

    async def can_retry(self, worker):
        """
        Check if the given worker is eligible to be sent to the retry queue.
        :param Worker worker: The resulting worker with exception
        :return bool:
        """
        backoff_context = await self._get_context(worker)
        conditions = await asyncio.gather(
            self._retry_count_is_lower(backoff_context),
            self._exception_in_list(worker.exception, backoff_context),
            self._interval_passed(backoff_context)
        )
        return all(conditions)

    async def update_context(self, context):
        # The arguments should be serializable....
        # We need to pop the context since it doesn't belong to the original
        # fn_kwargs
        retry_count = context['retry_count']
        context['retry_count'] = retry_count + 1
        context['last_retry'] = datetime.utcnow().timestamp()
        return context

    async def execute_retry(self, updated_worker, next_interval):
        """
        Hook method to implement the retry logic.

        This method should return an updated version of the worker (
        after the backoff was executed) or None for cases where the BackOff
        task will be sent to another service to handle. Ex: AsyncBackOff bellow.

        :param Worker updated_worker: A Worker with updated BackOff _meta
        :return Optional[Worker]: The Updated Worker with after backoff was
        done. If the backoff is asynchronous, returns None
        """
        raise NotImplementedError()

    async def retry(self, worker):
        """
        Called to invoke the retry action.
        It updates the backoff _meta attribute of the worker and send it
        :param Worker worker: the result worker with information to be used
        in the retry such as the fn_args and kwargs.
        """
        context = await self._get_context(worker)
        next_interval = await self.get_interval(context)
        meta = worker._meta or {}
        # Update the _meta context for the next retry before calling it
        # because of the async case.
        updated_context = await self.update_context(context)
        meta['backoff'] = updated_context
        worker._meta = meta  # meta might be a reference so this is ambiguous
        logger.info(f"{worker.target} failed, sending retry action.")
        return await self.execute_retry(worker, next_interval)

    async def bind(self, runner, parent=None):
        ext = await super(BackOff, self).bind(runner, parent=parent)
        if not isinstance(parent, Route):
            raise TypeError("A BackOff class can only be used inside a Route.")
        ext.route = parent
        return ext


class AsyncBackOff(BackOff, BaseEventListener):
    """
    BackOff that implements an Event Listener entrypoint which is responsible
    to manage the backoff of a specific route.

    Since it is an entrypoint, the binded route needs to be a copy, to enable
    us binding this entrypoint to it, instead of the original entrypoint.


    # TODO Maybe adding multiple entrypoints to a route would make this implementation cleaner
    """

    async def _interval_passed(self, context):
        return True  # RabbitMQ handles this.

    async def mount_backoff_payload(self, worker):
        return {
            'fn_args': worker.fn_args,
            'fn_kwargs': worker.fn_kwargs,
            '_meta': worker._meta
        }

    async def execute_retry(self, updated_worker, next_interval):
        # We fire the event to a queue known by the instantiated entrypoint
        retry_payload = await self.mount_backoff_payload(updated_worker)
        event_name = await self.get_event_name()
        exchange = await self.get_exchange()
        dispatch(
            self.runner.config["AMQP_URI"], exchange, event_name,
            retry_payload, headers={"x-delay": next_interval}
        )
        return None

    async def bind_route(self, route):
        self.route = route._clone()
        self.route._entrypoint = self
        await self.route.bind(self.runner)
        await self.route.entrypoint.bind_to_route(self.route)

    async def get_event_name(self):
        return f"{self.route.target}_backoff"

    async def get_exchange(self):
        """
        Return a different exchange due to the different type and arguments.
        :return Exchange:
        """
        return Exchange(
            name=self.service_name + '_backoff', type='x-delayed-message',
            arguments={'x-delayed-type': 'direct'}
        )

    async def get_queue(self):
        # TODO Multiple Routes to the same function will actually use the
        #  same queue here... is it okay?
        event_name = await self.get_event_name()
        exchange = await self.get_exchange()
        return Queue(
            name=f"{self.service_name}.{event_name}_queue",
            exchange=exchange, routing_key=event_name
        )

    async def new_entry(self, message, payload):
        # Converting the BackOff payload to the original payload and adding a
        # BackOff Context to it as kwargs
        fn_args = payload['fn_args']
        fn_kwargs = payload['fn_kwargs']
        _meta = payload['_meta']
        retry_count = _meta['backoff']['retry_count']
        max_retries = await self.get_max_retries(_meta["backoff"])
        logger.info(
            f"{self.route.target} retry {retry_count} of {max_retries}"
        )
        await self.call_route(message, *fn_args, _meta=_meta, **fn_kwargs)

    @property
    def service_name(self):
        return self.runner.config['SERVICE_NAME']



from micro_framework.rpc.dependencies import RPCDependency
from micro_framework.websocket.connectors import WebSocketConnector
from micro_framework.websocket.rpc import WSRPCProxy


class WebSocketRPCClient(RPCDependency):
    connector_class = WebSocketConnector
    proxy_class = WSRPCProxy

    def __init__(self, address, port, timeout=None):
        self.address = address
        self.port = port
        self.timeout = timeout

    def get_connector_kwargs(self):
        return {
            'uri': f"ws://{self.address}:{self.port}", 'timeout': self.timeout
        }



from prometheus_client import Gauge


unacked_rpc_messages = Gauge(
    'amqp_unacked_rpc_messages', 'Number of Pending RPC Messages',
    ["context", "route"]
)


unacked_event_messages = Gauge(
    'amqp_unacked_event_messages', 'Number of Pending Event Messages',
    ["context", "route"]
)



import asyncio
import time

import websockets

import micro_framework.amqp.amqp_elements
from micro_framework.exceptions import MaxConnectionsReached
from micro_framework.rpc import RPCTarget, parse_rpc_response, RPCProxy


class WSRPCTarget(RPCTarget):
    def call(self, message, client):
        connection_successful = False
        result = None
        while not connection_successful:
            # Try connecting to a server with available workers.
            # This is useful when the WS Entrypoint is behind a Load Balancer
            # and therefore we have multiple WebSocket servers, with some empty
            # and others not.

            connection = self.connector.get_connection()
            result_message = connection.send_and_receive(message)
            result = parse_rpc_response(result_message)
            if isinstance(result, MaxConnectionsReached):
                time.sleep(0.5)  # TODO Is there any better way?
                continue  # Keep Trying

            connection_successful = True
        return result


class WSRPCProxy(RPCProxy):
    target_class = WSRPCTarget



import asyncio
import logging
import uuid
from datetime import datetime
from timeit import default_timer as timer

from micro_framework.exceptions import PoolStopped, ExtensionIsStopped
from micro_framework.extensions import Extension
from micro_framework.routes import Route

logger = logging.getLogger(__name__)


class Entrypoint(Extension):
    route = None

    def bind(self, runner, parent=None):
        ext = super(Entrypoint, self).bind(runner, parent=parent)
        ext.route = parent
        return ext

    async def call_route(self, entry_id, *args, _meta=None, **kwargs):
        """
        Call the binded Route to start a new worker with the received content
        from the entrypoint.

        It returns the Instantiated Worker.

        This method caller should handle the following exceptions:

        . PoolStopped: Raises when the worker is signalled to stop.
        . Exception: Some internal error during the call to the worker.

        :param entry_id: Some object/identifier to be used to send a response
        for the correct caller.

        :param args: Additional arguments to be sent to the Route target
        function (such as payload for the event case)

        :param _meta: Metadata about this entrypoint, usually for internal
        use. (See AsyncRetry to an example)

        :param kwargs: Additional kwargs to be sent to the Route target
        function.
        """
        return await self.route.start_route(
            entry_id, *args, _meta=_meta, **kwargs
        )


class TimeEntrypoint(Entrypoint):
    def __init__(self, interval):
        self.interval = interval
        self.running = False

    async def _run_timer(self):
        last_time = timer()
        while self.running:
            elapsed_time = timer() - last_time
            sleep_time = self.interval - elapsed_time
            if sleep_time > 0:
                await asyncio.sleep(sleep_time)
            last_time = timer()
            entry_id = uuid.uuid4()
            now = datetime.utcnow()
            try:
                await self.call_route(entry_id, now.isoformat())
            except (ExtensionIsStopped, PoolStopped):
                pass

    async def start(self):
        """
        Start a loop that sleeps for a configured interval and then call the
        bind Route.

        Since it runs in a loop, some caution should be taken to prevent it
        from locking the event loop.
        """
        self.running = True
        # Since the runner call this using gather, we must run it with
        # ensure_future, so we don't stop the event-loop forever
        asyncio.ensure_future(self._run_timer())

    async def stop(self):
        self.running = False




import json
import uuid

from kombu import Queue

from micro_framework.amqp.amqp_elements import rpc_exchange, \
    rpc_routing_key, rpc_broadcast_routing_key, Publisher
from micro_framework.amqp.manager import RPCManager
from micro_framework.amqp.rpc import RPCReplyListener, _PicklableReplyListener
from micro_framework.rpc import RPCConnector, RPCConnection


class RPCProducer(Publisher, RPCConnection):
    """
    RPCProducer Implements the RPCConnection abstract class methods.

    At each "send" method, it will acquire a new connection from the
    producers pool (kombu library) and then send a message.

    It receives a send_to_listener callable, that is responsible to notify
    the EventListener of a new correlation_id to be aware of.
    """

    def __init__(
            self, amqp_uri, target_service: str,
            reply_listener: _PicklableReplyListener,
            **kwargs
    ):
        super(RPCProducer, self).__init__(amqp_uri, **kwargs)
        self.target_service = target_service
        self.reply_listener = reply_listener
        self.reply_to_queue = reply_listener.queue

    def send(self, payload, *args, **kwargs):
        payload = json.loads(payload)
        exchange = rpc_exchange()
        target_id = payload["command"]
        if target_id in RPCManager.internal_commands:
            routing_key = rpc_broadcast_routing_key(
                self.target_service, target_id
            )
        else:
            routing_key = rpc_routing_key(
                self.target_service, target_id=target_id
            )
        with self.reply_listener.producer_lock:
            # The publisher was raising some errors regarding wrong message
            # codes. Probably due to some concurrent call.
            # TODO This shouldn't have happened since we instantiate a new
            #  connection each time and use the get_producers internally....
            #  That uses the producers pool with acquire(block=True)
            self.publish(
                payload, routing_key=routing_key, exchange=exchange,
                **kwargs
            )

    def send_and_receive(self, *args, **kwargs):
        corr_id = str(uuid.uuid4())
        timeout = kwargs.pop("timeout", None)

        # Notify our reply listener of a new correlation_id that will have
        # ListenerReceiver.
        receiver = self.reply_listener.register_new_correlation(corr_id)
        self.send(
            *args, correlation_id=corr_id,
            reply_to=self.reply_to_queue.routing_key,
            mandatory=True,

            **kwargs
        )
        result = receiver.result(timeout=timeout)
        return result


class AMQPRPCConnector(RPCConnector):
    """

    Instantiate and returns a Publisher when get_connection method is called.

    The Arguments are:
        .amqp_uri: The Connection String to the broker.

        .reply_to_queue: The queue to which we will tell the RPCServer to
        send the response message.
    """

    def __init__(
            self, amqp_uri: str, target_service: str,
            reply_listener: _PicklableReplyListener
    ):
        self.amqp_uri = amqp_uri
        self.target_service = target_service
        self.reply_listener = reply_listener

    def get_connection(self):
        return RPCProducer(
            self.amqp_uri, reply_listener=self.reply_listener,
            target_service=self.target_service
        )



class FrameworkException(Exception):
    pass


class PoolStopped(Exception):
    pass

class ExtensionIsStopped(FrameworkException):
    pass


class RPCTargetDoesNotExist(FrameworkException):
    pass


class RPCException(FrameworkException):
    def __init__(self, exception_dict):
        self.exc_type = exception_dict['exception']
        self.value = exception_dict['message']
        if self.value:
            message = f"{self.exc_type} {self.value}"
        else:
            message = self.exc_type
        super(RPCException, self).__init__(message)


class MaxConnectionsReached(FrameworkException):
    pass


class RPCAsyncResponse(FrameworkException):
    """
    Exception used to differentiate when a RPC call should return or not

    We raise an exception instead of returning None because None could be the
    desired response.
    """

class BrokenSpawner(FrameworkException):
    ...



import logging
from functools import partial
from typing import Dict

from kombu import Exchange, producers

from micro_framework.amqp.amqp_elements import get_connection, Publisher
from micro_framework.amqp.connectors import AMQPRPCConnector
from micro_framework.amqp.rpc import RPCReplyListener
from micro_framework.dependencies import Dependency, RunnerDependency
from micro_framework.rpc import RPCDependency

logger = logging.getLogger(__name__)


class Producer(Dependency):
    def __init__(self, confirm_publish=True, service_name=None):
        self.service_name = service_name
        self.confirm_publish = confirm_publish

    @property
    def exchange(self):
        service_name = self.service_name or self.config['SERVICE_NAME']
        exchange_name = f"{service_name}.events"  # Nameko Compatible
        return Exchange(exchange_name, type='topic', auto_delete=True)

    @property
    def amqp_uri(self):
        return self.config['AMQP_URI']

    @property
    def transport_options(self):
        return {
            'confirm_publish': self.confirm_publish
        }
    
    def bind(self, runner, parent=None):
        return super(Producer, self).bind(runner, parent)

    async def setup(self):
        self._publisher = Publisher(self.amqp_uri)

    async def get_dependency(self, worker):
        self.config = worker.config
        exchange = self.exchange

        def dispatch_event(event_name, payload):
            return self._publisher.publish(
                payload, exchange=exchange, routing_key=event_name
            )

        return dispatch_event


class RPCProxyProvider(RPCDependency):
    """
    Provides a RPCProxy with AMQPRPCConnector.
    """

    def __init__(self, target_service):
        self.target_service = target_service
        super(RPCProxyProvider, self).__init__()

    reply_listener = RPCReplyListener()
    connector_class = AMQPRPCConnector

    def get_connector_kwargs(self) -> Dict:
        return {
            "amqp_uri": self.config.get("AMQP_URI"),
            "target_service": self.target_service,
            "reply_listener": self.reply_listener.picklable_listener,
        }


def dispatch(amqp_uri, exchange, routing_key, payload, **kwargs):
    """
    Helper to dispatch a single payload that will load the client and call
    the send method.

    :param str amqp_uri:
    :param Exchange exchange:
    :param str routing_key:
    :param Any payload:
    :param dict headers:
    """
    connection = get_connection(amqp_uri)
    with producers[connection].acquire(block=True) as producer:
        return producer.publish(
            payload, exchange=exchange, routing_key=routing_key,
            **kwargs
        )



import asyncio
import inspect
from functools import partial
from typing import Union, Callable, Dict

from micro_framework.dependencies import Dependency, RunnerDependency
from micro_framework.extensions import Extension


async def call_dependencies(
        dependencies: Dict, action: str, *args, **kwargs) -> Dict:
    futures = []
    names = []
    for name, dependency in dependencies.items():
        method = getattr(dependency, action)
        futures.append(method(*args, **kwargs))
        names.append(name)

    results = await asyncio.gather(*futures)
    ret = {name: result for name, result in zip(names, results)}
    return ret


class AsyncTargetExecutor:
    def __init__(self, config, target, dependencies, runner_dependencies):
        self._config = config
        self.dependencies = dependencies
        self.runner_dependencies = runner_dependencies or {}
        self.target = target

    async def pre_call(self):
        await call_dependencies(
            self.dependencies, "setup_dependency", self
        )
        self.injected_dependencies = await call_dependencies(
            self.dependencies, "get_dependency", self
        )

    async def after_call(self):
        await call_dependencies(
            self.dependencies, "after_call", self, self.result,
            self.exception
        )
        if self.exception:
            raise self.exception

    async def run(self, *args, **kwargs):
        """
        Run the async target function/method.

        :return : Returns the target result
        :raises: An Exception raised from the target can be raised.
        """
        await self.pre_call()
        kwargs.update(**self.runner_dependencies) # Already loaded from runner
        kwargs.update(**self.injected_dependencies)

        self.exception = None
        self.result = None
        try:
            self.result = await self.target(*args, **kwargs)
        except Exception as exc:
            self.exception = exc
        finally:
            await self.after_call()
        return self.result

    @property
    def config(self):
        return self._config


class TargetExecutor(AsyncTargetExecutor):
    def run(self, *args, **kwargs):
        """
        Run the target function/method synchronously.

        This method is called when the worker has an executor to call the
        target.

        If the target is a coroutine and the worker_mode is asyncio,
        then run_async should be called instead.

        :return : Returns the target result
        :raises: An Exception raised from the target can be raised.
        """
        event_loop = asyncio.get_event_loop()
        asyncio.run_coroutine_threadsafe(
            self.pre_call(),
            event_loop
        ).result()

        kwargs.update(**self.runner_dependencies)
        kwargs.update(**self.injected_dependencies)

        self.exception = None
        self.result = None
        try:
            self.result = self.target(*args, **kwargs)
        except Exception as exc:
            self.exception = exc
        finally:
            asyncio.run_coroutine_threadsafe(
                self.after_call(),
                event_loop
            ).result()

        return self.result



class ClassExecutorMixin:
    def __init__(self, config, class_instance, target_method,
                 class_dependencies, target_runner_dependencies,
                 target_dependencies):
        super(ClassExecutorMixin, self).__init__(
            config, target_method, target_runner_dependencies,
            target_dependencies
        )
        self.class_instance = class_instance
        self.class_dependencies = class_dependencies

    async def pre_call(self):
        await call_dependencies(
            self.class_dependencies, "setup_dependency", self
        )
        injected_cls_dependencies = await call_dependencies(
            self.class_dependencies, "get_dependency", self
        )
        for name, dep in injected_cls_dependencies.items():
            setattr(self.class_instance, name, dep)
        return await super(ClassExecutorMixin, self).pre_call()

    async def after_call(self):
        await call_dependencies(
            self.class_dependencies, "after_call", self, self.result,
            self.exception
        )
        return await super(ClassExecutorMixin, self).after_call()


class AsyncClassTargetExecutor(ClassExecutorMixin, AsyncTargetExecutor):
    ...


class ClassTargetExecutor(ClassExecutorMixin, TargetExecutor):
    ...


class Target(Extension):
    """
    Represents a Function to be called by a Worker.

    This Target controls the Dependencies that are going to be called and
    injected.

    After all setup, the worker will call its mount_target method, passing the
    worker instance. This method will return a TargetExecutor with the target
    RunnerDependencies injected and the Dependencies to be injected.

    """
    def __init__(self, target):
        self._runner_dependencies = {}
        self.all_dependencies = {}
        self._target = target
        self.populate_dependencies()

    def get_target_executor_class(self):
        if inspect.iscoroutinefunction(self.target):
            return AsyncTargetExecutor
        return TargetExecutor

    def populate_dependencies(self):
        self.all_dependencies = getattr(self.target, "__dependencies__", {})
        self._dependencies = {}
        self._runner_dependencies = {}
        for name, dependency in self.all_dependencies.items():
            if isinstance(dependency, RunnerDependency):
                # RunnerDependencies are handled outside the worker.
                self._runner_dependencies[name] = dependency
            else:
                # Dependencies are handled inside the worker
                self._dependencies[name] = dependency

        self._target = getattr(self.target, "__original_function", self.target)

    async def call_dependencies(
            self, dependencies: Dict, action: str, *args, **kwargs) -> Dict:
        futures = []
        names = []
        for name, dependency in dependencies.items():
            method = getattr(dependency, action)
            futures.append(method(*args, **kwargs))
            names.append(name)

        results = await asyncio.gather(*futures)
        ret = {name: result for name, result in zip(names, results)}
        return ret

    async def on_worker_setup(self, worker):
        return await self.call_dependencies(
            self._runner_dependencies, "setup_dependency", worker
        )

    async def mount_target(self, worker):
        """
        Prepare this target by returning a TargetExecutor.
        """
        dependencies = await self.call_dependencies(
            self._runner_dependencies, "get_dependency", worker
        )
        target_executor_class = self.get_target_executor_class()
        executor_instance = target_executor_class(
            self.runner.config, self.target,
            self._dependencies, dependencies
        )
        return executor_instance

    async def on_worker_finished(self, worker):
        return await self.call_dependencies(
            self._runner_dependencies, "after_call", worker, worker.result,
            worker.exception
        )

    async def target_detail(self):
        """
        Return the Details of an entrypoint's target.
        :param Entrypoint entrypoint: Entrypoint
        :return dict: Entrpoint Details dict.
        """
        target_id = str(self)

        target_parameters = self.parameters

        args = [
            name for name, parameter in target_parameters.items() if
            parameter.kind in [
                inspect.Parameter.POSITIONAL_OR_KEYWORD,
                inspect.Parameter.POSITIONAL_ONLY,
                inspect.Parameter.VAR_POSITIONAL
            ] and parameter.default is inspect._empty  # arguments with
            # default are considered POSITIONAL_OR_KEYWORD. So here I'll
            # filter only those that don't have a default value.
        ]
        kwargs = [
            name for name, parameter in target_parameters.items() if
            parameter.kind in [
                inspect.Parameter.KEYWORD_ONLY, inspect.Parameter.VAR_KEYWORD,
                inspect.Parameter.POSITIONAL_OR_KEYWORD
            ] and (parameter.default is not inspect._empty or
                   parameter.kind is inspect.Parameter.VAR_KEYWORD)
        ]
        return {
            'target': target_id,
            'args': args,
            'kwargs': kwargs,
            'docstring': inspect.getdoc(self.target)
        }

    @property
    def target(self) -> Callable:
        return self._target

    @property
    def parameters(self):
        """
        Returns the Parameters that the method accepts. (Excluding the
        Injected Dependencies)
        :return:
        """
        return {
            name: param
            for name, param in inspect.signature(self.target).parameters.items()
            if name not in self.all_dependencies
        }

    def __str__(self):
        return self.target.__name__

    def __repr__(self):
        return "<Target {}>".format(self.target.__name__)


class TargetFunction(Target):
    def bind(self, runner, parent=None):
        ext = super(TargetFunction, self).bind(runner, parent=parent)
        for name, dependency in ext._dependencies.items():
            extension = ext.bind_new_extension(
                name, dependency, runner, ext
            )
            ext._dependencies[name] = extension

        for name, dependency in ext._runner_dependencies.items():
            extension = ext.bind_new_extension(
                name, dependency, runner, ext
            )
            ext._runner_dependencies[name] = extension

        return ext


class TargetClassMethod(Target):
    """
    Represents a class to be instanced and then have a specific method
    be called by a Worker.

    It lists all dependencies to be injected into the class and into the method.
    """

    def __init__(self, target_class, target_method):
        self.target_class = target_class
        self.target_method = target_method
        method = getattr(target_class, target_method)
        super(TargetClassMethod, self).__init__(method)
        # The Superclass will handle all method handling, here we will extend
        # it to the class.
        self._class_runner_dependencies = {}
        self.populate_class_dependencies()

    def get_target_executor_class(self):
        if inspect.iscoroutinefunction(self.target):
            return AsyncClassTargetExecutor
        return ClassTargetExecutor

    def populate_class_dependencies(self):
        dependencies = inspect.getmembers(
            self.target_class, lambda x: isinstance(x, Dependency)
        )
        self._class_dependencies = {}
        for name, dependency in dependencies:
            if isinstance(dependency, RunnerDependency):
                self._class_runner_dependencies[name] = dependency
            else:
                self._class_dependencies[name] = dependency

    def bind(self, runner, parent=None):
        ext = super(TargetClassMethod, self).bind(runner, parent=parent)

        for name, dependency in ext._class_dependencies.items():
            extension = ext.bind_new_extension(
                name, dependency, runner, ext
            )
            ext._class_dependencies[name] = extension

        for name, dependency in ext._class_runner_dependencies.items():
            extension = ext.bind_new_extension(
                name, dependency, runner, ext
            )
            ext._class_runner_dependencies[name] = extension

        return ext

    async def on_worker_setup(self, worker):
        await super(TargetClassMethod, self).on_worker_setup(worker)
        return await self.call_dependencies(
            self._class_runner_dependencies, "setup_dependency", worker
        )

    async def mount_target(self, worker):
        cls = await self.get_class_instance(worker)
        dependencies = await self.call_dependencies(
            self._runner_dependencies, "get_dependency", worker
        )
        target_executor_class = self.get_target_executor_class()
        return target_executor_class(
            self.runner.config, cls, getattr(cls, self.target_method),
            self._class_dependencies, dependencies,
            self._dependencies
        )

    async def on_worker_finished(self, worker):
        await super(TargetClassMethod, self).on_worker_finished(worker)
        return await self.call_dependencies(
            self._class_runner_dependencies, "after_call", worker,
            worker.result, worker.exception
        )

    async def get_class_instance(self, worker):
        injected_cls_runner_dependencies = await self.call_dependencies(
            self._class_runner_dependencies, "get_dependency", worker
        )
        cls = self.target_class()
        for name, dependency in injected_cls_runner_dependencies.items():
            setattr(cls, name, dependency)
        return cls

    def __str__(self):
        return "{}.{}".format(
            self.target_class.__name__, self.target_method
        )

    def __repr__(self):
        return "<TargetClassMethod {}.{}>".format(
            self.target_class.__name__, self.target_method
        )


def new_target(target: Callable, method_name=None) -> Union[
    TargetFunction, TargetClassMethod]:
    """
    Instantiates a Target based on the parameters given.

    If the target parameter is a class, then we return a TargetClassMethod
    class.

    Else we return a TargetFunction class.

    :param callable target: A Function or Class
    :param str method_name: If a class is given, then this will be the method
    to be called.
    :return Union[TargetFunction, TargetClassMethod]: The Instantiated Target.
    """
    if inspect.isclass(target) and not method_name:
        method_name = "__call__"
    elif not inspect.isclass(target):
        method_name = None

    if method_name:
        return TargetClassMethod(target, method_name)

    return TargetFunction(target)



# TODO get configurations somehow
import asyncio
import inspect
import logging.config
import sys
import time
from concurrent.futures._base import CancelledError

from functools import partial
from timeit import default_timer


from micro_framework.config import FrameworkConfig
from micro_framework.entrypoints import Entrypoint
from micro_framework.exceptions import ExtensionIsStopped, BrokenSpawner
from micro_framework.extensions import Extension
from micro_framework.metrics import PrometheusMetricServer, tasks_counter, \
    tasks_number, tasks_failures_counter, tasks_duration, info, \
    available_workers
from micro_framework.routes import Route
from micro_framework.spawners import SPAWNERS
from micro_framework.spawners.thread import ThreadSpawner



logger = logging.getLogger(__name__)


class RunnerContext:
    """
    It is a context of the framework, where the routes will share the same
    spawners (and therefore the max_workers number)
    """
    def __init__(
            self, routes=None, worker_mode=None, max_workers=None,
            context_metric_label=None
    ):
        if routes is None:
            routes = []

        self.routes = []
        self.config = {}
        if max_workers:
            self.config["MAX_WORKERS"] = max_workers

        if worker_mode:
            self.config["WORKER_MODE"] = worker_mode

        self.is_running = False
        self.spawner = None
        self.spawned_extensions = {}
        self.entrypoints = set()
        self.extra_extensions = set()
        self.context_singletons = {} # {type: instance}
        self.metric_label = context_metric_label or f"context_{id(self)}"
        self.running_routes = 0
        self.executor = None
        self.add_routes(*routes)

    def add_routes(self, *routes):
        self.routes += routes
        return self

    async def bind(self, runner):
        self.runner = runner

    async def setup(self):
        # Update this context config with the runner config to fill missing
        # configurations.
        self.config = FrameworkConfig(
            self.config, default_config=self.runner.config
        )
        self.executor = SPAWNERS["greedy_thread"]({})
        self.metric_server = self.runner.metric_server

        await self.bind_routes()
        logger.debug(f"Extensions: {self.extensions}")

        info.labels(self.metric_label).info(
            {
                'max_workers': str(self.config["MAX_WORKERS"]),
                'worker_mode': self.config["WORKER_MODE"],
            }
        )
        available_workers.labels(self.metric_label).set(self.available_workers)

    async def _call_extensions_action(self, action, extension_set=None):
        if extension_set is None:
            extension_set = self.extensions
        tasks = [getattr(extension, action)() for extension in extension_set]
        await asyncio.gather(*tasks)

    async def _find_extensions(self):
        for route in self.routes:
            extensions = await find_extensions(route)
            for extension in extensions:
                await self.register_extension(extension)

    @property
    def extensions(self):
        return {*self.routes, *self.entrypoints, *self.extra_extensions}

    async def bind_routes(self):
        binded_routes = []
        for route in self.routes:
            if not isinstance(route, Route):
                raise TypeError("Only Route class should be added as a "
                                "route.")
            binded_routes.append(route.bind(self))
        self.routes = binded_routes

    def register_extension(self, extension):
        if isinstance(extension, Entrypoint) and not extension.singleton:
            self.entrypoints.add(extension)
        else:
            self.extra_extensions.add(extension)

        if extension.context_singleton:
            existing = self.context_singletons.get(type(extension))
            if existing is not None and existing != extension:
                raise TypeError(
                    "Extension {} is a context singleton and already has a "
                    "registered Extension: {}".format(
                        extension, existing
                    )
                )
            self.context_singletons[type(extension)] = extension
        if extension.singleton:
           self.runner.register_extension(extension)

    async def start(self):
        self.event_loop = asyncio.get_event_loop()
        logger.info(
            f"Starting Runner with {self.config['MAX_WORKERS']}"
            f" {self.config['WORKER_MODE']} workers"
        )

        logger.debug(f"Extensions: {self.extensions}")

        # Once we have all extensions detected:
        # Setup Routes first then entrypoints and then extra extensions.
        await self._call_extensions_action('setup', extension_set=self.routes)
        await self._call_extensions_action(
            'setup', extension_set=self.entrypoints
        )
        await self._call_extensions_action(
            'setup', extension_set=self.extra_extensions
        )

        # Finally, we start all extensions.
        self.is_running = True
        await self._call_extensions_action('start')
        self.spawner = await self.runner.get_spawner(self)

    async def stop(self, wait=True):
        self.is_running = False

        logger.info("Stopping all context extensions and workers")

        # Stopping Routes
        await self._call_extensions_action('stop', extension_set=self.routes)

        if self.spawner:
            # Stop Workers
            logger.debug("Stopping workers")
            self.spawner.stop(wait=wait)

        logger.debug("Stopping extra extensions")
        await self._call_extensions_action(
            'stop', extension_set=self.extra_extensions
        )

        # Stop running extensions
        logger.debug("Stopping any still running extensions thread")
        # self.extension_spawner.stop(wait=False)
        # Summary: Please Stop!!
        for task in self.spawned_extensions:
            task.cancel()

    async def spawn_worker(
            self, route, worker, callback=None
    ):
        """
        Spawn a new worker to execute a function from the calling route.

        :param Route route: The route that called it
        :param fn_args: Arguments of the function to be called.
        :return Future: A Future object of the task.
        """
        if not self.is_running:
            raise ExtensionIsStopped()

        # Update metrics before starting the worker.
        tasks_counter.labels(self.metric_label, route.metric_label).inc()
        tasks_number.labels(self.metric_label, route.metric_label).inc()
        available_workers.labels(self.metric_label).dec()
        start = default_timer()

        # Run the worker until it completes the task

        # Note that since we are working with asyncio, we don't need to
        # create a task, the await will automatically giveup on the lock and
        # only return when the task in the worker is done.
        self.running_routes += 1
        try:
            await worker.run(self)
        except BrokenSpawner:
            # Unrecoverable error. It's better to simply stop everything
            # without even acknowledging messages.
            await self.runner._stop(wait=False)
            raise
        self.running_routes -= 1

        # Update the metrics after the worker is finished.
        tasks_number.labels(self.metric_label, route.metric_label).dec()
        available_workers.labels(self.metric_label).inc()
        if worker.exception is not None:
            tasks_failures_counter.labels(
                self.metric_label, route.metric_label
            ).inc()
        duration = default_timer() - start
        tasks_duration.labels(
            self.metric_label, route.metric_label
        ).observe(duration)

        return worker

    def execute_async(self, coroutine, raise_for_future=True):
        """
        Executes a coroutine in the background, without awaiting it.

        This method is to normal functions/methods from extensions to call an
        async function.

        :param coroutine: Coroutine
        :return Future: Future object.
        """
        future = asyncio.run_coroutine_threadsafe(coroutine, self.event_loop)
        def on_finished_task(future):
            if future.exception() and raise_for_future:
                raise future.exception()
        future.add_done_callback(on_finished_task)
        return future

    async def sync_to_async(self, fn, *args, **kwargs):
        fn = partial(fn, *args, **kwargs)
        return await self.event_loop.run_in_executor(self.executor, fn)

    async def spawn_extension(
            self, extension, target_fn, *fn_args, **fn_kwargs
    ):
        """
        Any extension that needs an async call should call this method to
        start it.

        :param Extension extension: An extension subclass
        :param function target_fn: the function to be executed.
        :param function callback: A callback function to handle the
        result/error
        :param fn_args: Any argument to be passed to the target function.
        :param fn_kwargs: Any named argument to be passed to the target
        function.
        """
        if inspect.iscoroutinefunction(target_fn):
            coroutine = target_fn(*fn_args, **fn_kwargs)
            try:
                await coroutine
            except (CancelledError, asyncio.CancelledError):
                return

        else:  # Sync Functions will run in the default executor
            target_fn = partial(target_fn, *fn_args, **fn_kwargs)
            future = self.event_loop.run_in_executor(None, target_fn)

            future.add_done_callback(self.extension_thread_callback)
            self.spawned_extensions[future] = extension

    def extension_thread_callback(self, future):
        """
        Handle the result of an extension when it's worker is finished either
        by an error or simply shutting down.
        """
        if not future.cancelled() and future.exception() and not isinstance(
                future.exception(), (CancelledError, asyncio.CancelledError)):
            raise future.exception()

    @property
    def max_workers(self):
        return self.config["MAX_WORKERS"]

    @property
    def available_workers(self):
        return self.max_workers - self.running_routes

    @property
    def singletons(self):
        return self.runner.singletons


class Runner:
    def __init__(self, contexts=None, config=None, metric_server=None):
        if contexts is None:
            contexts = []

        if config is None:
            config = {}

        self.spawners = []
        self.config = FrameworkConfig(config)
        if contexts and isinstance(contexts[0], Route):
            contexts = [RunnerContext(contexts)]
        self.contexts = contexts
        self.metric_server = metric_server
        self.extensions = set()
        self.singletons = {}

    def exception_handler(self, loop, context):
        asyncio.run_coroutine_threadsafe(self._stop(), self.event_loop)
        if context.get("exception"):
            raise context.get("exception")

    def add_contexts(self, *contexts):
        self.contexts += contexts

    def new_context(
            self, routes=None, worker_mode=None, max_workers=None,
            context_metric_label=None, inplace=False):
        ctx = RunnerContext(routes, worker_mode, max_workers, context_metric_label)
        if inplace:
            self.add_contexts(ctx)
        return ctx

    def register_extension(self, extension):
        if extension.singleton:
            existing = self.singletons.get(type(extension))
            if existing is not None and existing != extension:
                raise TypeError(
                    "Extension {} is a singleton and already has a registered "
                    "Extension: {}".format(extension, existing)
                )
            self.singletons[type(extension)] = extension
        self.extensions.add(extension)

    async def get_spawner(self, runner_context: RunnerContext):
        """
        Returns and register a new spawner instance.
        :param str spawner_type: The type of the Spawner
        :param spawner_configs: That spawner configurations
        :return Spawner: A spawner instance.
        """
        if runner_context.config["WORKER_MODE"] == "asyncio":
            return None

        spawner = SPAWNERS[runner_context.config["WORKER_MODE"]](
            runner_context.config
        )
        self.spawners.append(spawner)
        return spawner

    async def _start(self):
        # Metric Extension Binding
        self.event_loop = asyncio.get_event_loop()

        if self.metric_server is None:
            self.metric_server = PrometheusMetricServer()
        self.metric_server = self.metric_server.bind(self)

        for context in self.contexts:
            await context.bind(self)
            await context.setup()
        await self.metric_server.setup()
        await self.metric_server.start()


        for context in self.contexts:
            await context.start()

    async def _stop(self, wait=True):
        # TODO Ensure that all spawner pools are stopped so then we stop the
        #  rest
        cancel_contexts = [context.stop(wait=wait) for context in self.contexts]
        await asyncio.gather(*cancel_contexts, return_exceptions=True)

        if sys.version_info[1] < 7:
            tasks = [
                t for t in asyncio.Task.all_tasks(self.event_loop)
                if t is not asyncio.Task.current_task(self.event_loop)
            ]
        else:
            tasks = [
                t for t in asyncio.all_tasks() if t is not asyncio.current_task()
            ]
        for task in tasks: task.cancel()
        await asyncio.gather(*tasks, return_exceptions=True)
        await self.metric_server.stop()
        self.event_loop.call_soon_threadsafe(self.event_loop.stop)
        logger.info("Runner stopped")


    def start(self):
        self.event_loop = asyncio.get_event_loop()
        self.event_loop.set_exception_handler(self.exception_handler)
        try:
            self.event_loop.run_until_complete(self._start())
            self.event_loop.run_forever()
        except (KeyboardInterrupt, asyncio.CancelledError):
            self.event_loop.run_until_complete(self._stop())
        except Exception:
            self.event_loop.run_until_complete(self._stop())
            raise


async def find_extensions(cls: Extension, existing_extensions=None):
    """
    Searches for extensions inside a given class.


    :param cls: Any class
    :param set existing_extensions: A set of extensions already registered.
    :return set: Set of extensions found
    """
    if existing_extensions is None:
        existing_extensions = set()
    extensions = existing_extensions

    for extension in cls._extensions:
        if extension in existing_extensions:
            continue
        extensions.add(extension)
        extensions.update(await find_extensions(extension, extensions))
    return extensions



from .connectors import RPCConnector, RPCConnection
from .dependencies import RPCDependency, RPCDependencyMixin
from .proxies import RPCTarget, RPCProxy
from .manager import RPCManagerMixin
from .formatters import parse_rpc_response, format_rpc_command, format_rpc_response

__all__ = [
    "RPCConnector", "RPCConnection", "RPCDependency",
    "RPCDependencyMixin", "RPCManagerMixin", "format_rpc_command",
    "parse_rpc_response", "format_rpc_response"
]



import threading
from concurrent.futures._base import Future
from typing import Dict, List, Set, Any, Type

from .connectors import RPCConnector
from .formatters import parse_rpc_response, format_rpc_command


class RPCTarget:
    """
    Represents a Route's target from the connected server.

    It exposes methods to call the target synchronously or asynchronously.

    """

    def __init__(self, connector: RPCConnector, target, args=None, kwargs=None,
                 docstring=None):
        self.connector = connector
        self.target = target
        self.target_args = args
        self.target_kwargs = kwargs
        self.target_doc = docstring

    def call(self, message: str, connector: RPCConnector) -> Dict:
        connection = connector.get_connection()
        result_message = connection.send_and_receive(message)
        return parse_rpc_response(result_message)

    def _call_client_task(
            self, message: str, connector: RPCConnector, future: Future):
        result = self.call(message, connector)
        if isinstance(result, Exception):
            future.set_exception(result)
            return
        future.set_result(result)

    def call_async(self, *args, **kwargs) -> Future:
        """
        Start a new thread to call the target.

        It returns a Future object with result/done methods.

        :return Future: Future object with result
        """

        message = format_rpc_command(self.target, *args, **kwargs)
        future = Future()

        # TODO Create an executor instance in the worker context (specially
        #  for the instantiated worker).
        t = threading.Thread(
            target=self._call_client_task,
            args=(message, self.connector, future)
        )
        t.daemon = True
        t.start()

        return future

    @property
    def info(self):
        return f"""
RPC Target Object.
> Target: {self.target}
> Target's Arguments: {self.target_args}
> Target's Key Arguments: {self.target_kwargs}
> Target's Docstring: \n{self.target_doc}
        """

    def __call__(self, *args, **kwargs) -> Any:
        future = self.call_async(*args, **kwargs)
        return future.result()

    def __repr__(self):
        return f"RPC<{self.target}>"

    def __str__(self):
        return self.__repr__()


class RPCProxy:
    """
    Injected Dependency that populates itself based on the list of targets
    from the connected RPC server.

    The class object attributes are the RPC Server targets instantiated as a
    RPCTarget class object.

    """
    target_class = RPCTarget

    def __init__(self, connector: RPCConnector, targets: List[Dict],
                 name: str = None):
        self._targets = set()
        self.connector = connector
        self.name = name

        for target in targets:
            self.add_target(target)

    def get_target_class(self) -> Type[RPCTarget]:
        return self.target_class

    def get_target_kws(self, **target) -> Dict:
        return {
            "connector": self.connector,
            **target,
        }

    def get_target(self, **target) -> RPCTarget:
        return self.get_target_class()(**self.get_target_kws(**target))

    def new_nested_proxy(self, name: str) -> 'RPCProxy':
        return self.__class__(self.connector, [], name=name)

    def add_target(self, target: dict, nested_name: str = None) -> RPCTarget:
        target_name = nested_name or target['target']
        target_name, *nested = target_name.split('.', 1)
        if nested:
            obj = getattr(self, target_name, self.new_nested_proxy(target_name))
            obj.add_target(target, nested[0])
        else:
            obj = self.get_target(**target)
        setattr(self, target_name, obj)
        self._targets.add(obj)
        return obj

    @property
    def targets(self) -> Set[RPCTarget]:
        return self._targets

    def __repr__(self):
        name = self.name or 'Server'
        return f"RPC<{name}>"

    def __str__(self):
        return self.__repr__()



import setuptools

setuptools.setup(
    packages=setuptools.find_packages(exclude=["tests"]),
    name='micro-framework',
    url='https://github.com/Mendes11/micro_framework',
    version='2.0.14',
    description='Framework to create CPU or IO bound microservices.',
    long_description= 'file: README.md',
    author = 'Rafael Mendes Pacini Bachiega',
    author_email = 'rafaelmpb11@hotmail.com',
    classifiers =
    ['Intended Audience :: Developers',
    'Operating System :: OS Independent',
    'Programming Language :: Python :: 3 :: Only',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
    'Programming Language :: Python :: 3.9'],
    install_requires=[
        "kombu>=4.6.8,<5", "websockets>=8.1,<9", "websocket-client>=0.57.0,<1",
        "prometheus-client<0.9", "loky>=2.9.0,<3", "psutil"
    ],
)



import inspect
import json

import micro_framework
from micro_framework.exceptions import RPCException


def import_exception(exception):
    fw_exceptions = dict(inspect.getmembers(
        micro_framework.exceptions,
        lambda x: inspect.isclass(x) and issubclass(x, Exception)
    ))
    if exception['exception'] in __builtins__:
        return __builtins__[exception['exception']](exception['message'])
    elif exception["exception"] in fw_exceptions:
        return fw_exceptions[exception["exception"]](exception.get("message"))
    return RPCException(exception)


def format_rpc_command(command, *args, **kwargs):
    """
    Returns a json str with the expected structure to communicate with our
    RPCManager.

    :param str command: RPC Command. Can be either list/get targets or a call
    to the target by its name.
    :param args: Arguments to be sent to the RPC Command.
    :param kwargs: Key Arguments to be sent to the RPC Command.
    :return str: JSON Command
    """
    return json.dumps({
        'command': command, 'command_args': args, 'command_kwargs': kwargs
    })


def format_rpc_response(data, exception=None):
    """
    Formats a response from a RPC Manager.
    It provides the data and/or a serialized exception so it can be
    re-created by the caller.

    :param Any data: A JSON Serializable object.
    :param Exception exception: An Exception object
    :return str: JSON Response.
    """
    exception_data = None
    if exception:
        exception_data = {
            'exception': type(exception).__name__,
            'message': str(exception),
        }
    return json.dumps({
        'data': data,
        'exception': exception_data
    })


def parse_rpc_response(message):
    """
    Reads a response returned by the RPC Manager.
    It returns either a loaded JSON or an exception if the exception key is
    not None.

    :param str message: JSON Response
    :return dict|list|Exception: Loaded RPC Response.
    """
    response = json.loads(message)
    if response.get('exception'):
        return import_exception(response['exception'])
    return response.get('data')


import asyncio
import logging
import time
from multiprocessing import Pipe, Manager

from micro_framework.amqp.amqp_elements import rpc_reply_queue, rpc_exchange, \
    get_connection
from micro_framework.amqp.entrypoints import BaseEventListener

logger = logging.getLogger(__name__)


class ListenerReceiver:
    """
    Implements a method to receive from the Reply EventListener's connection
    instance with a timeout possibility.

    When the timeout is reached, it raises TimeoutError.

    This class is usually created by the RPCConnection when it sends a new
    rpc message to the broker and notify the Reply-Listener to send the reply
    to a specific multiprocess.Connection instance.

    """

    def __init__(self, connection):
        self.connection = connection

    def result(self, timeout=None):
        if timeout is not None:
            if self.connection.poll(timeout=timeout):
                return self.connection.recv()
            raise TimeoutError()
        while not self.connection.poll(timeout=.5):
            pass
        return self.connection.recv()


class _CorrelationResultIntent:
    def __init__(self, correlation_id, sender):
        self.correlation_id = correlation_id
        self.sender = sender

    def reply_to_sender(self, reply):
        if not self.sender.closed and self.sender.writable:
            return self.sender.send(reply)
        raise ValueError("Reply sender is closed.")


class RPCReplyListener(BaseEventListener):
    """
    Listens to any RPCReply for this service
    """
    singleton = True
    internal = True

    async def setup(self):
        self.mp_manager = Manager()

        self._reply_queue = rpc_reply_queue()
        self.routing_key = self._reply_queue.routing_key
        self.running = False

        self.producer_lock = self.mp_manager.Lock()
        self.replies_lock = self.mp_manager.Lock()

        # This is a shared variable between all processes. It maps the
        # correlation ids and their intents to be returned to the user.
        self.reply_intents = self.mp_manager.dict()
        self.replies = {}
        await super(RPCReplyListener, self).setup()

    async def start(self):
        self.running = True
        await self.runner.spawn_extension(
            self, self.reply_sender
        )
        await super(RPCReplyListener, self).start()

    async def stop(self):
        """
        At stop, we delete the Reply Queue because it is not needed anymore.
        (Because it is unique per running service and therefore,
        if we restart the service a new queue name will be generated).

        """
        await self.manager.stop()
        self.running = False
        self.mp_manager.shutdown()
        # We need to wait 1 second due to the manager's ConsumerMixin timeout
        # to identify that the Consumer should stop. Otherwise an exception
        # will be generated because of the Consumer losing the queue reference.
        await asyncio.sleep(1)
        conn = get_connection(self.runner.config["AMQP_URI"])
        conn.connect()
        self._reply_queue.bind(conn.channel())
        self._reply_queue.delete()
        conn.close()
        await super(RPCReplyListener, self).stop()

    async def get_exchange(self):
        return rpc_exchange()

    async def get_queue(self):
        return self._reply_queue

    def reply_sender(self):
        while self.running:
            if not self.replies:
                time.sleep(.01)
                continue

            with self.replies_lock:
                replies = list(self.replies.items())
            for corr_id, reply in replies:
                intent = self.reply_intents.get(corr_id)
                if not intent:
                    continue
                intent.reply_to_sender(reply)

                with self.replies_lock:
                    self.replies.pop(corr_id)
                    self.reply_intents.pop(corr_id)

            time.sleep(.01)

    @property
    def picklable_listener(self):
        return _PicklableReplyListener(self)

    async def call_route(self, entry_id, *args, _meta=None, **kwargs):
        # Any exception here will make the manager requeue the message
        # First we update our correlation_ids dict by emptying the
        # correlations_queue.
        # await self.read_correlations_queue()
        try:
            corr_id = entry_id.properties["correlation_id"]
        except KeyError:
            logger.error("Received a reply with no registered correlation_id")
            return

        body = args[0]
        with self.replies_lock:
            self.replies[corr_id] = body


class _PicklableReplyListener:
    """
    A version of the ReplyListener to be sent to worker processes.
    """

    def __init__(self, reply_listener: RPCReplyListener):
        self.producer_lock = reply_listener.producer_lock
        self.replies_lock = reply_listener.replies_lock
        self.reply_intents = reply_listener.reply_intents
        self.queue = reply_listener.queue

    def register_new_correlation(self, correlation_id):
        receiver, sender = Pipe(duplex=False)
        with self.replies_lock:
            self.reply_intents[correlation_id] = _CorrelationResultIntent(
                correlation_id, sender
            )
        return ListenerReceiver(receiver)



import logging

from micro_framework.exceptions import ExtensionIsStopped
from micro_framework.extensions import Extension
from micro_framework.targets import new_target
from micro_framework.workers import Worker, CallbackWorker

logger = logging.getLogger(__name__)


class Route(Extension):
    """
    Base Extension Destined to link an endpoint to an entry function.

    Protocol-Specific routes should implement this class and extend its
    features.

    """
    worker_class = Worker

    def __init__(
            self, target, entrypoint, translators=None,
            worker_class=None, backoff=None, method_name=None,
            metric_label=None):
        self.target = new_target(target, method_name=method_name)
        self._entrypoint = entrypoint
        self._translators = translators or []
        self._method_name = method_name
        self.stopped = False
        self.current_workers = {}
        self.worker_class = worker_class or self.worker_class
        self._backoff = backoff
        self._metric_label = metric_label

    async def handle_finished_worker(self, entry_id, worker):
        """
        Override this method to do some validations after a worker has
        finished.

        This method should return either the worker instance or None.

        :param entry_id:
        :param worker:
        :return:
        """
        return worker

    async def run_backoff(self, entry_id, worker):
        """
        Run a backoff class.

        This method will return the worker from the retry.
        If None is returned, then it indicates that the backoff is running
        asynchronously and we let the entrypoint decide what to do.

        :param entry_id: Some object/identifier to be used to send a response
        for the correct caller.

        :param Worker worker: Worker Instance with result/exception attrs

        :return Optional[Worker]: Worker Instance with result/exception
        attrs or a None type, indicating an asynchronous backoff was made.
        """
        return await self.backoff.retry(worker)

    async def get_worker_instance(self, *fn_args, _meta=None, **fn_kwargs):
        return self.worker_class(
            self.target, self.translators.copy(), self.runner.config,
            *fn_args, _meta=_meta, **fn_kwargs
        )

    async def run_worker(self, entry_id, worker):
        logger.debug(f"{self} is Spawning worker {worker}")
        worker = await self.runner.spawn_worker(self, worker)
        logger.debug(
            f"{self} Received a worker result: {worker.result}."
        )
        if worker.exception:
            can_retry = self.backoff and await self.backoff.can_retry(worker)
            if can_retry:
                worker = await self.run_backoff(entry_id, worker)

        return await self.handle_finished_worker(entry_id, worker)

    async def start_route(self, entry_id, *fn_args, _meta=None, **fn_kwargs):
        if self.stopped:
            raise ExtensionIsStopped()
        worker = await self.get_worker_instance(
            *fn_args, _meta=_meta, **fn_kwargs
        )
        return await self.run_worker(entry_id, worker)

    async def stop(self):
        self.stopped = True

    @property
    def translators(self):
        return self._translators

    @property
    def entrypoint(self):
        return self._entrypoint

    @property
    def backoff(self):
        return self._backoff

    @property
    def metric_label(self):
        return self._metric_label or self.__str__().replace(" -> ", "__")

    def __str__(self):
        target = str(self.target)
        return f'{self.__class__.__name__} -> {target}'

    def __repr__(self):
        target = str(self.target)
        return "< {} -> {} >".format(self.__class__.__name__, target)


class CallbackRoute(Route):
    # FIXME This don't work anymore. Refactor with the target classes.
    callback_worker_class = CallbackWorker

    def __init__(self, *args, callback_target,
                 callback_worker_class=None, **kwargs):
        super(CallbackRoute, self).__init__(*args, **kwargs)
        self.callback_target = callback_target
        self.callback_worker_class = callback_worker_class or self.callback_worker_class

    async def get_callback_worker_instance(self, original_worker):
        return self.callback_worker_class(
            self.callback_target, original_worker
        )

    async def start_callback_route(self, entry_id, worker):
        logger.info(f"{self.target} failed. Starting callback route.")
        callback_worker = await self.get_callback_worker_instance(worker)
        return await self.run_worker(entry_id, callback_worker)

    async def handle_finished_worker(self, entry_id, worker):
        # Callback route will only be called if retry is not valid.
        if worker is None:
            return worker

        if worker.exception and self.callback_target:
            logger.debug(
                "Finished worker has an exception, calling callback route."
            )
            worker = await self.start_callback_route(entry_id, worker)

        return super(CallbackRoute, self).handle_finished_worker(entry_id, worker)



import logging

import websockets

logger = logging.getLogger(__name__)


class WebSocketServer:
    """
    Asyncio WebSocket Server.

    It is started by calling the start_server method and passing a running
    event_loop.

    """

    def __init__(self):
        self.clients = set()

    async def register_client(self, websocket):
        logger.debug(f"Client {websocket} connected.")
        self.clients.add(websocket)

    async def unregister_client(self, websocket):
        logger.debug(f"Client {websocket} disconnected.")
        self.clients.remove(websocket)

    async def listener(self, websocket, path):
        logger.debug(f"Received new Client: {websocket}.")
        await self.register_client(websocket)

        try:
            async for message in websocket:
                logger.debug(f"New message from {websocket}: {message}")
                await self.message_received(websocket, message)

        finally:
            await self.unregister_client(websocket)

    async def send(self, websocket, message):
        if not websocket.closed:
            await websocket.send(message)

    async def message_received(self, websocket, message):
        pass

    async def serve(self, ip, port):
        self.server = await websockets.serve(self.listener, ip, port)
        logger.info(f"WebSocketManager listening at {ip}:{port}")
        return self.server

    async def stop(self):
        if self.server:
            self.server.close()

    def start_server(self, event_loop, ip, port):
        """
        Starts the WebSocket Server inside the event_loop.
        :param event_loop:
        :param str ip: IP address to bind the server at.
        :param port: Port to bind the server at.
        """
        event_loop.run_until_complete(self.serve(ip, port))




import inspect
import json
import logging

from micro_framework.exceptions import RPCTargetDoesNotExist
from micro_framework.extensions import Extension
from .formatters import format_rpc_response
from ..entrypoints import Entrypoint

logger = logging.getLogger(__name__)

async def target_detail(entrypoint: Entrypoint):
    """
    Return the Details of an entrypoint's target.
    :param Entrypoint entrypoint: Entrypoint
    :return dict: Entrpoint Details dict.
    """
    target = entrypoint.route.target
    target_id = str(target)

    target_parameters = target.parameters

    args = [
        name for name, parameter in target_parameters.items() if
        parameter.kind in [
            inspect.Parameter.POSITIONAL_OR_KEYWORD,
            inspect.Parameter.POSITIONAL_ONLY,
            inspect.Parameter.VAR_POSITIONAL
        ] and parameter.default is inspect._empty  # arguments with
        # default are considered POSITIONAL_OR_KEYWORD. So here I'll
        # filter only those that don't have a default value.
    ]
    kwargs = [
        name for name, parameter in target_parameters.items() if
        parameter.kind in [
            inspect.Parameter.KEYWORD_ONLY, inspect.Parameter.VAR_KEYWORD,
            inspect.Parameter.POSITIONAL_OR_KEYWORD
        ] and (parameter.default is not inspect._empty or
               parameter.kind is inspect.Parameter.VAR_KEYWORD)
    ]
    return {
        'target': target_id,
        'args': args,
        'kwargs': kwargs,
        'docstring': inspect.getdoc(target)
    }


class RPCRegister:
    def __init__(self):
        self._targets = {}

    async def add_entrypoint(self, entrypoint):
        target = await target_detail(entrypoint)
        self._targets[target["target"]] = entrypoint

    async def remove_entrypoint(self, entrypoint):
        target = await target_detail(entrypoint)
        self._targets.pop(target["target"], None)

    @property
    def entrypoints(self):
        return list(self.targets.values())

    @property
    def targets(self):
        return self._targets


class RPCManagerMixin(Extension):
    """
    Manages RPC messages to call Entrypoints

    Messages using this Manager should have the following json schema:

    {
        "command": str,
        "command_args": list
    }


    Available Commands are:

     - __list_targets__
        Lists all possible targets with their parameters and docstring.
        - args: None
        - kwargs: None

     - __get_target__:
        Retrieve detail from a specific target
        - args: target str
        - kwargs: None

     - <target>
        Call the specified target, triggering the Route
        - args: Arguments for the specific target
        - kwargs: Kwargs for the specific target.

    """

    def __init__(self):
        super(RPCManagerMixin, self).__init__()
        self.entrypoints = {}

    context_singleton = True
    internal_commands = {"__list_targets__", "__get_target__"}
    register = RPCRegister()  # This is unique amongst all RPCManagers.

    async def add_entrypoint(self, entrypoint):
        await self.register.add_entrypoint(entrypoint)
        detail = await target_detail(entrypoint)
        target_id = detail['target']
        if target_id in self.entrypoints:
            raise ValueError(
                f"Entrypoint {entrypoint} is already registered. "
                f"You cannot have multiple entrypoints to the same target"
            )
        self.entrypoints[target_id] = entrypoint

    async def _parse_message(self, client, command, *args, **kwargs):
        try:
            method = getattr(self, command)
            response, exception = await method(client, *args, **kwargs)
        except AttributeError:
            response, exception = await self.__call_target__(
                client, command, *args, **kwargs
            )
        return response, exception

    async def consume_message(self, client, message):
        """
        Consume the message, returning the method based on the rpc command.

        If the command is to call a route, then it should return None,
        since the route will spawn a worker, running in parallel.

        :param client: Client to identify the caller
        :param str message: RPC Message
        :return str: RPC Response
        """
        if not isinstance(message, dict):
            try:
                message = json.loads(message)
            except json.JSONDecodeError as exc:
                logger.error(
                    "RPCManager received unknown message structure: {}".format(
                        message
                    )
                )
                return format_rpc_response(None, exc)

        command = message['command']
        command_args = message.get('command_args', [])
        command_kwargs = message.get('command_kwargs', {})

        try:
            response, exception = await self._parse_message(
                client, command, *command_args, **command_kwargs
            )
        except Exception as exc:
            response = None
            exception = exc

        return format_rpc_response(response, exception)

    async def call_entrypoint(self, client, entrypoint, *args, **kwargs):
        return await entrypoint.call_route(client, *args, **kwargs)

    async def __list_targets__(self, client):
        """
        List all available targets by iterating through the entrypoints in
        the register.
        :param client: RPC Client
        :return list, Optional[Exception]: All available Entrypoints's target
        """
        return [
                   await target_detail(entrypoint)
                   for entrypoint in self.register.entrypoints
               ], None

    async def __get_target__(self, client, target):
        """
        Return details from an specific entrypoint target.
        :param client: RPC Client
        :param target: target rpc identification
        :return dict, Optional[Exception]: Target detail
        """
        return await target_detail(self.register.targets[target]), None

    async def __call_target__(self, client, target, *args, **kwargs):
        """
        Call the entrypoint's route by using the target name.
        :param client: RPC Client
        :param str target: Entrypoint Identification
        :param dict data:
        :return:
        """
        if target not in self.entrypoints:
            raise RPCTargetDoesNotExist(
                f"{target} does not exist or is not exposed through a "
                f"RPC method."
            )
        entrypoint = self.entrypoints[target]
        worker = await self.call_entrypoint(
            client, entrypoint, *args, **kwargs
        )

        data = worker.result
        exception = worker.exception
        return data, exception



import json


class MessageTranslator:
    """
    Simple class receives a message and returns a transformation of it.
    """
    async def get_message(self, *args, **kwargs):
        return args[0]

    async def set_message(self, message, *args, **kwargs):
        list(args)[0] = message
        return args, kwargs

    async def translate_message(self, message):
        raise NotImplementedError()

    async def translate(self, *args, **kwargs):
        message = await self.get_message(*args, **kwargs)
        translated = await self.translate_message(message)
        return await self.set_message(translated, *args, **kwargs)


class JSONTranslator(MessageTranslator):
    async def translate_message(self, message):
        if isinstance(message, dict):
            return message
        return json.loads(message)
